 /*
  * 
 * <LIC_AMD_STD>
 * Copyright (c) 2004 Advanced Micro Devices, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * The full GNU General Public License is included in this distribution in the
 * file called COPYING
 * </LIC_AMD_STD>
  * 
  * <CTL_AMD_STD>
  * </CTL_AMD_STD>
  * 
  * <DOC_AMD_STD>
  * </DOC_AMD_STD>
  * 
  */

#include "gxlnxaud.h"

PDURAUDIO	pGeode;
static struct nsc_card *devs = NULL;

static int geode_pm_callback(struct pm_dev *dev, pm_request_t rqst, void *d);

static struct pci_driver geode_pci_driver;

static int	use_pm=1;	
static int  gAllocatedIOResources = FALSE;

MODULE_AUTHOR		("Mario Raposo");
MODULE_DESCRIPTION	("NSC Geode Audio Driver");
MODULE_LICENSE		("GPL");

#ifdef M_DEBUG
	MODULE_PARM		(debug,"i");
#endif

MODULE_PARM		(use_pm,	"i");
MODULE_PARM_DESC	(use_pm,
	"Set to 0 to disable power management. Default is 1");

#include "build_num.h"
#define DRIVER_VERSION "1.2." GEODEOSS_BUILD_NUM

#define PCI_VENDOR_CYRIX				0x1078
#define PCI_VENDOR_NSC					0x100B		
	
#define PCI_DEVICE_ID_NSC_GX1			0x0103		// GX1	
#define PCI_DEVICE_ID_NSC_SCXX00     	0x0503		// SCXX00
#define PCI_DEVICE_ID_NSC_GX2	     	0x002E		// GX2	

#define NSC_CHAN_HARD		0x100

#define NSC_FMT_STEREO		0x01
#define NSC_FMT_16BIT		0x02
#define NSC_FMT_MASK		0x03
#define NSC_DAC_SHIFT		0   
#define NSC_ADC_SHIFT		4

#define DAC_RUNNING			1
#define ADC_RUNNING			2
							
#define MAX_DSP_ORDER		2
#define MAX_DSPS			(1<<MAX_DSP_ORDER)
#define NR_IDRS				32
							
#define NR_APUS				64
#define NR_APU_REGS			16

//
// ACPI states 
//
enum 
{
	ACPI_D0=0,
	ACPI_D1,
	ACPI_D2,
	ACPI_D3
};

//
// Bits in the acpi masks 
//
#define ACPI_12MHZ	( 1 << 15)
#define ACPI_24MHZ	( 1 << 14)
#define ACPI_978	( 1 << 13)
#define ACPI_SPDIF	( 1 << 12)
#define ACPI_GLUE	( 1 << 11)
#define ACPI__10	( 1 << 10) // reserved 
#define ACPI_PCIINT	( 1 << 9)
#define ACPI_HV		( 1 << 8) // hardware volume 
#define ACPI_GPIO	( 1 << 7)
#define ACPI_ASSP	( 1 << 6)
#define ACPI_SB		( 1 << 5) // sb emul 
#define ACPI_FM		( 1 << 4) // fm emul 
#define ACPI_RB		( 1 << 3) // ringbus / aclink 
#define ACPI_MIDI	( 1 << 2) 
#define ACPI_GP		( 1 << 1) // game port 
#define ACPI_WP		( 1 << 0) // wave processor 

#define ACPI_ALL	(0xffff)
#define ACPI_SLEEP	(~(ACPI_SPDIF|ACPI_ASSP|ACPI_SB|ACPI_FM| \
			ACPI_MIDI|ACPI_GP|ACPI_WP))
#define ACPI_NONE	(ACPI__10)

//
//  these masks indicate which units we care about at
//	which states 
//
u16 acpi_state_mask[] = {
	[ACPI_D0] = ACPI_ALL,
	[ACPI_D1] = ACPI_SLEEP,
	[ACPI_D2] = ACPI_SLEEP,
	[ACPI_D3] = ACPI_NONE
};

static char version[] __devinitdata =
KERN_INFO "geode OSS: version " DRIVER_VERSION " time " __TIME__ " " __DATE__ "\n";

static const unsigned sample_size[]		= { 1, 2, 2, 4 };
static const unsigned sample_shift[]	= { 0, 1, 1, 2 };

enum card_types_t 
{
	TYPE_GX1,
	TYPE_SCXX00,
	TYPE_GX2
};

static const char *card_names[]=
{
	[TYPE_GX1]		= "GX1 OSS",
	[TYPE_SCXX00]	= "SCXX00 OSS",
	[TYPE_GX2]		= "GX2 OSS"
};

static int geode_notifier(struct notifier_block *nb, unsigned long event, void *buf);

static struct notifier_block geode_nb = {geode_notifier, NULL, 0};

// --------------------------------------------------------------------- 

struct nsc_state 
{
	u8	apu		[6];		// l/r output, l/r intput converters, l/r input apus 
	u8	apu_mode[6];		// Running mode for this APU 
	u8	apu_pan	[6];		// Panning setup for this APU 
	u32 apu_base[6];		// base address for this apu 

	struct nsc_card *card;	// Card info 

	//
	// wave stuff 
	//
	unsigned int rateadc, ratedac;
	unsigned char fmt, enable;

	int index;

	//
	// Only let 1 be opening at a time 
	//
	struct semaphore open_sem;
	wait_queue_head_t open_wait;
	mode_t open_mode;

	//
	// soundcore stuff 
	//
	int dev_audio;

	struct dmabuf 
	{
		void *rawbuf;
		unsigned buforder;
		unsigned numfrag;
		unsigned fragshift;

		//
		// XXX zab - swptr only in here so that it can be referenced by
		// 			 clear_advance, as far as I can tell :( 
		//
		unsigned	hwptr, swptr;
		unsigned	total_bytes;
		int			count;
		unsigned	error; // over/underrun 

		wait_queue_head_t wait;

		//
		// redundant, but makes calculations easier 
		//
		unsigned 	fragsize;
		unsigned 	dmasize;
		unsigned 	fragsamples;

		unsigned	char	BitsPerSample;
		unsigned	char	nChannels;
		volatile    long	BytesReadyInDMABuffer;
		//
		//	Flags
		//
		char		zero_flags; // flags to control underrun zeroing
		unsigned	DMAmapped;
		unsigned	ready;		// our oss buffers are ready to go 

		unsigned	ossfragsize;
		unsigned	ossfragshift;
		int			ossmaxfrags;

		unsigned	subdivision;

		struct		vm_area_struct vma;			//  Used to save the mapping structure (For DirectDMA)

		u16 base;					// Offset for ptr 
	} dma_dac, dma_adc;

	//
	// pointer to each dsp?s piece of the apu->src buffer page 
	//
	void *mixbuf;
};

//
// define the meaning for the zero_flags field
#define ZF_TRANSFER_PENDING    1
#define ZF_TRANSFER_SINCE_ZERO 2
#define ZF_PARTIAL_ZERO        3
#define ZF_ZERO_SKIPPED        4

// The threshhold for zeroing the buffer is a the interrupt interval + 1ms
// Worst case 1ms is 48KHz 16 bit stereo i.e. 48 samples of 4 bytes each
#define Z_1MS_OF_BYTES ( 48 * 4 )
	
struct nsc_card 
{
	unsigned int magic;

	//
	// We keep geode cards in a linked list 
	//
	struct nsc_card *next;

	int dev_mixer;

	int card_type;

	//
	//  As most of this is static,
	//	perhaps it should be a pointer to a global struct 
	//
	struct mixer_goo 
	{
		int modcnt;
		int supported_mixers;
		int stereo_mixers;
		int record_sources;
//		int (*read_mixer)(struct nsc_card *card, int index);
		void (*write_mixer)(struct nsc_card *card,int mixer, unsigned int left,unsigned int right);
		int (*recmask_io)(struct nsc_card *card,int rw,int mask);
		unsigned int mixer_state[SOUND_MIXER_NRDEVICES];
	} mix;
	
	int power_regs;
		
	int in_suspend;
	wait_queue_head_t suspend_queue;

	struct nsc_state channels[MAX_DSPS];
	u16 geode_map[NR_IDRS];					// Register map 

	//
	// We have to store this junk so that we can come back from a
	//	suspend 
	//
	u16 apu_map[NR_APUS][NR_APU_REGS];		// contents of apu regs 

	//
	// this locks around the physical registers on the card 
	//
	spinlock_t lock;

	//
	// memory for this card.. wavecache limited 
	//
	void *dmapages;
	int dmaorder;

	//
	//. hardware resources 
	//
	struct pci_dev *pcidev;
	u32 iobase;
	u32 irq;

	int dock_mute_vol;
};

static void set_mixer(struct nsc_card *card,unsigned int mixer, unsigned int val );

//
//	AC97 codec programming interface.
//
	 
static void geode_ac97_set(struct nsc_card *card, u8 cmd, u16 val)
{
	OS_DbgMsg("geodeoss: ac97_set %x -> %x\n",cmd,val);
	DURAUDIO_CodecWrite(pGeode,cmd,val);
}

static u16 geode_ac97_get(struct nsc_card *card, u8 cmd)
{
	return DURAUDIO_CodecRead(pGeode, cmd);
}

//
// OSS interface to the ac97s.. 
//

#define AC97_STEREO_MASK (SOUND_MASK_VOLUME|\
	SOUND_MASK_PCM|SOUND_MASK_LINE|SOUND_MASK_CD|\
	SOUND_MASK_VIDEO|SOUND_MASK_LINE1|SOUND_MASK_IGAIN)

#define AC97_SUPPORTED_MASK (AC97_STEREO_MASK | \
	SOUND_MASK_BASS|SOUND_MASK_TREBLE|SOUND_MASK_MIC|\
	SOUND_MASK_SPEAKER)

#define AC97_RECORD_MASK (SOUND_MASK_MIC|\
	SOUND_MASK_CD| SOUND_MASK_VIDEO| SOUND_MASK_LINE1| SOUND_MASK_LINE|\
	SOUND_MASK_PHONEIN)

#define supported_mixer(CARD,FOO) ( CARD->mix.supported_mixers & (1<<FOO) )

//
//  this table has default mixer values for all OSS mixers.
//	be sure to fill it in if you add oss mixers
//	to anyone's supported mixer defines 
//

#define PACKED_U8(hi,lo) ( ( (hi & 0xFF) << 8 ) | ( lo & 0xFF ) )
#define VOL_DEFAULT PACKED_U8(50,50)

// The mixer default values are in 0-100 OSS units
static unsigned int mixer_defaults[SOUND_MIXER_NRDEVICES] = {
	[SOUND_MIXER_VOLUME] =          VOL_DEFAULT,
	[SOUND_MIXER_BASS] =            VOL_DEFAULT,
	[SOUND_MIXER_TREBLE] =          VOL_DEFAULT,
	[SOUND_MIXER_SPEAKER] =         VOL_DEFAULT,
	[SOUND_MIXER_MIC] =				0x8000, // NOTE: the mic is different.
	[SOUND_MIXER_LINE] =			VOL_DEFAULT,
	[SOUND_MIXER_CD] =				VOL_DEFAULT,
	[SOUND_MIXER_VIDEO] =			VOL_DEFAULT,
	[SOUND_MIXER_LINE1] =			VOL_DEFAULT,
	[SOUND_MIXER_PCM] =             VOL_DEFAULT,
	[SOUND_MIXER_IGAIN] =           VOL_DEFAULT
};
	
// The AC97 values are in AC97 control register units
static struct ac97_mixer_hw {
	unsigned char offset;
	int scale;
} ac97_hw[SOUND_MIXER_NRDEVICES]= {
	[SOUND_MIXER_VOLUME]	=	{MASTER_VOLUME,MASTER_VOLUME_MAX},
	[SOUND_MIXER_BASS]		=	{MASTER_TONE_RL,15},
	[SOUND_MIXER_TREBLE]	=	{MASTER_TONE_RL,15},
	[SOUND_MIXER_SPEAKER]	=	{PC_BEEP_VOLUME,15},
	[SOUND_MIXER_MIC]		=	{MIC_VOLUME,31},
	[SOUND_MIXER_LINE]		=	{LINE_IN_VOLUME,31},
	[SOUND_MIXER_CD]		=	{CD_VOLUME,31},
	[SOUND_MIXER_VIDEO]		=	{VIDEO_VOLUME,31},
	[SOUND_MIXER_LINE1]		=	{TV_VOLUME,31},
	[SOUND_MIXER_PCM]		=	{PCM_OUT_VOL,31},
	[SOUND_MIXER_IGAIN]		=	{RECORD_GAIN,15}
};

#if 0 // *shrug* removed simply because we never used it.
	  // feel free to implement again if needed */

//
// reads the given OSS mixer from the ac97
//	the caller must have insured that the ac97 knows
//	about that given mixer, and should be holding a
//	spinlock for the card 
//
static int ac97_read_mixer(struct nsc_card *card, int mixer) 
{
	u16 val;
	int ret=0;
	struct ac97_mixer_hw *mh = &ac97_hw[mixer];

	val = geode_ac97_get(card, mh->offset);

	if(AC97_STEREO_MASK & (1<<mixer)) {
		// nice stereo mixers .. 
		int left,right;

		left = (val >> 8)  & 0x7f;
		right = val  & 0x7f;

		if (mixer == SOUND_MIXER_IGAIN) {
			right = (right * 100) / mh->scale;
			left = (left * 100) / mh->scale;
		else {
			right = 100 - ((right * 100) / mh->scale);
			left = 100 - ((left * 100) / mh->scale);
		}

		ret = left | (right << 8);
	} else if (mixer == SOUND_MIXER_SPEAKER) {
		ret = 100 - ((((val & 0x1e)>>1) * 100) / mh->scale);
	} else if (mixer == SOUND_MIXER_MIC) {
		ret = 100 - (((val & 0x1f) * 100) / mh->scale);
	//  the low bit is optional in the tone sliders and masking
	//	it lets is avoid the 0xf 'bypass'.. 
	} else if (mixer == SOUND_MIXER_BASS) {
		ret = 100 - ((((val >> 8) & 0xe) * 100) / mh->scale);
	} else if (mixer == SOUND_MIXER_TREBLE) {
		ret = 100 - (((val & 0xe) * 100) / mh->scale);
	}

	OS_DbgMsg("read mixer %d (0x%x) %x -> %x\n",mixer,mh->offset,val,ret);

	return ret;
}
#endif

// write the OSS encoded volume to the given OSS encoded mixer,
//	again caller's job to make sure all is well in arg land,
//	call with spinlock held 
	
// linear scale -> log 
static unsigned char lin2log[101] = 
{
0, 0 , 15 , 23 , 30 , 34 , 38 , 42 , 45 , 47 ,
50 , 52 , 53 , 55 , 57 , 58 , 60 , 61 , 62 ,
63 , 65 , 66 , 67 , 68 , 69 , 69 , 70 , 71 ,
72 , 73 , 73 , 74 , 75 , 75 , 76 , 77 , 77 ,
78 , 78 , 79 , 80 , 80 , 81 , 81 , 82 , 82 ,
83 , 83 , 84 , 84 , 84 , 85 , 85 , 86 , 86 ,
87 , 87 , 87 , 88 , 88 , 88 , 89 , 89 , 89 ,
90 , 90 , 90 , 91 , 91 , 91 , 92 , 92 , 92 ,
93 , 93 , 93 , 94 , 94 , 94 , 94 , 95 , 95 ,
95 , 95 , 96 , 96 , 96 , 96 , 97 , 97 , 97 ,
97 , 98 , 98 , 98 , 98 , 99 , 99 , 99 , 99 , 99 
};

static void ac97_write_mixer(struct nsc_card *card,int mixer, unsigned int left, unsigned int right)
{
	u16 val=0;
	struct ac97_mixer_hw *mh = &ac97_hw[mixer];

	if(AC97_STEREO_MASK & (1<<mixer)) 
	{
		//
		// stereo mixers, mute them if we can 
		//

		if (mixer == SOUND_MIXER_IGAIN) 
		{
			//
			// igain's slider is reversed.. 
			//
			right = (right * mh->scale) / 100;
			left = (left * mh->scale) / 100;
			if ((left == 0) && (right == 0))
				val |= MUTE_MASK;
		} 
		else 
		{
			//
			// log conversion for the stereo controls 
			//
			if((left == 0) && (right == 0))
				val = MUTE_MASK;

			right = ((100 - lin2log[right]) * mh->scale) / 100;
			left = ((100 - lin2log[left]) * mh->scale) / 100;
		}

		val |= (left << 8) | right;
	} 
	else 
		if (mixer == SOUND_MIXER_SPEAKER) 
		{
			val = (((100 - left) * mh->scale) / 100) << 1;
		} 
		else 
			if (mixer == SOUND_MIXER_MIC) 
			{
				val = geode_ac97_get(card, mh->offset) & ~0x801f;
				val |= (((100 - left) * mh->scale) / 100);
				//
				//  the low bit is optional in the tone sliders and masking
				//	it lets is avoid the 0xf 'bypass'.. 
				//
			} 
			else 
				if (mixer == SOUND_MIXER_BASS) 
				{
					val = geode_ac97_get(card , mh->offset) & ~0x0f00;
					val |= ((((100 - left) * mh->scale) / 100) << 8) & 0x0e00;
				} 
				else 
					if (mixer == SOUND_MIXER_TREBLE)  
					{
						val = geode_ac97_get(card , mh->offset) & ~0x000f;
						val |= (((100 - left) * mh->scale) / 100) & 0x000e;
					}

	geode_ac97_set(card , mh->offset, val);

	if ( MASTER_VOLUME == mh->offset ) {
#ifdef GEODE_HPOUT_IS_LINELVL
		// If HPOUT is being treated as a line out, only track the mute
		// state of the master volume
		u16 mute = val & MUTE_MASK
		if ( MUTE_MASK != mute )
			val = LINE_LEV_RESET_VOL & ~MUTE_MASK;
		else
			val = LINE_LEV_RESET_VOL | MUTE_MASK;
#endif
		// Since AC97 CODEC have two independent "master volumes"
		// and OSS doesn't, slave the HPOUT aka LINE_LEV to the
		// master volume
		geode_ac97_set(card , LINE_LEV_OUT_VOL, val);
	}

	
}

//
// the following tables allow us to go from 
//	OSS <-> ac97 quickly. 
//

enum ac97_recsettings 
{
	AC97_REC_MIC=0,
	AC97_REC_CD,
	AC97_REC_VIDEO,
	AC97_REC_AUX,
	AC97_REC_LINE,
	AC97_REC_STEREO,	// combination of all enabled outputs..  
	AC97_REC_MONO,      // or the mono equivalent 
	AC97_REC_PHONE        
};

static unsigned int ac97_oss_mask[] = 
{
	[AC97_REC_MIC]		= SOUND_MASK_MIC, 
	[AC97_REC_CD]		= SOUND_MASK_CD, 
	[AC97_REC_VIDEO]	= SOUND_MASK_VIDEO, 
	[AC97_REC_AUX]		= SOUND_MASK_LINE1, 
	[AC97_REC_LINE]		= SOUND_MASK_LINE, 
	[AC97_REC_PHONE]	= SOUND_MASK_PHONEIN
};

//
// indexed by bit position 
//
static unsigned int ac97_oss_rm[] = 
{
	[SOUND_MIXER_MIC]		= AC97_REC_MIC,
	[SOUND_MIXER_CD]		= AC97_REC_CD,
	[SOUND_MIXER_VIDEO]		= AC97_REC_VIDEO,
	[SOUND_MIXER_LINE1]		= AC97_REC_AUX,
	[SOUND_MIXER_LINE]		= AC97_REC_LINE,
	[SOUND_MIXER_PHONEIN]	= AC97_REC_PHONE
};
	
//
//  Read or write the recmask 
//	the ac97 can really have left and right recording
//	inputs independently set, but OSS doesn't seem to 
//	want us to express that to the user. 
//	the caller guarantees that we have a supported bit set,
//	and they must be holding the card's spinlock 
//
static int ac97_recmask_io(struct nsc_card *card, int read, int mask) 
{
	unsigned int val = ac97_oss_mask[ geode_ac97_get(card, 0x1a) & 0x7 ];

	if (read) 
		return val;

	//
	//  OSS can have many inputs, the geode cant.  try
	//	to pick the 'new' one 
	//
	if (mask != val) 
		mask &= ~val;

	val = ffs(mask) - 1; 
	val = ac97_oss_rm[val];
	val |= val << 8;			// set both channels 

	OS_DbgMsg("geodeoss: setting ac97 recmask to 0x%x\n",val);

	geode_ac97_set(card,0x1a,val);

	return 0;
};

//
//	Indirect register access. Not all registers are readable so we
//	need to keep register state ourselves
//
 
#define WRITEABLE_MAP	0xEFFFFF
#define READABLE_MAP	0x64003F

//
//	These routines handle accessing the second level indirections to the
//	wave ram.
//
static void set_dac_rate(struct nsc_state *s, unsigned int rate)
{
	OS_DbgMsg("Setting DAC rate to: %d\n", rate);

	s->ratedac = rate;
	DURAUDIO_SetCodecRate(pGeode, CHANNEL0_PLAYBACK, (unsigned long) rate);
}

static void set_adc_rate(struct nsc_state *s, unsigned int rate)
{
	OS_DbgMsg("Setting ADC rate to: %d\n", rate);

	s->rateadc = rate;
	DURAUDIO_SetCodecRate(pGeode, CHANNEL1_RECORD, (unsigned long) rate);
}

static inline void stop_adc(struct nsc_state *s)
{
	if (! (s->enable & ADC_RUNNING)) 
		return;
 
	s->enable &= ~ADC_RUNNING;

	DURAUDIO_StopDMA(pGeode,CHANNEL1_RECORD);
}	

static void stop_dac(struct nsc_state *s)
{
	OS_DbgMsg("stop_dac\n");

	if (! (s->enable & DAC_RUNNING)) 
	{
		OS_DbgMsg("DAC Not running! - Returning...\n");
		return;
	}
	s->enable &= ~DAC_RUNNING;

	DURAUDIO_StopDMA(pGeode,CHANNEL0_PLAYBACK);
}	

#define START_DAC_DEFAULT   0x0
#define START_DAC_START_ANY 0x1


//
// define the meanin for the zero_flags field
#define ZF_TRANSFER_PENDING    1
#define ZF_TRANSFER_SINCE_ZERO 2
#define ZF_PARTIAL_ZERO        3

// Zero the unused bytes in the playback dma buffer
void zero_free_playback_buffer(struct nsc_state *s)
{
	OS_DbgMsg("geodeoss: zero_free_playback_buffer\n");
	clear_bit(ZF_TRANSFER_SINCE_ZERO,&s->dma_dac.zero_flags);
	DURAUDIO_ZeroFillFreeBuffer(pGeode);
}

static void start_dac(struct nsc_state *s, int flags)
{
	long ready  =  s->dma_dac.BytesReadyInDMABuffer;
	long thresh =  s->dma_dac.fragsize;

	if ( s->enable & DAC_RUNNING ) {
		OS_DbgMsg("start_dac: DAC Already running. Returning...\n");
		return;
	}

	OS_DbgMsg("start_dac: bytes ready: %d  start threshhold: %d \n",
		ready, thresh);
	//
	//	We only start if we have at least a fragment
	//	or we have ANY bytes and are told to ignore the threshhold
	//
	if ( (ready >= thresh) ||
	     (ready && (flags & START_DAC_START_ANY ) ) )
	{
		OS_DbgMsg ("start_dac: Starting DMA\n");

		// Before we start playback, make sure the rest of the
		// buffer zero'd for underrun
		zero_free_playback_buffer(s);
		s->enable |= DAC_RUNNING;
		DURAUDIO_StartDMA(pGeode,CHANNEL0_PLAYBACK);
	}
}	

// Utility to get bytes in use safely and update cached value
unsigned long get_playback_bytes_used(struct nsc_state *s)
{
	unsigned long used;
	if ( s->enable & DAC_RUNNING) 
		// If the DAC is running, the ISR updates the value
		used = s->dma_dac.BytesReadyInDMABuffer;
	else  {
		// if not, we get it from Durango, and update the local copy
		used = DURAUDIO_UpdateByteCounter (pGeode, CHANNEL0_PLAYBACK); 
		s->dma_dac.BytesReadyInDMABuffer = used;
	}
	return used;
}

static void start_adc(struct nsc_state *s)
{
	s->enable |= ADC_RUNNING;

	DURAUDIO_StartDMA(pGeode,CHANNEL1_RECORD);
}
	
static void nsc_close_playback (struct nsc_state *s)
{
	stop_dac(s);

	OS_DbgMsg ("Closing Stream. \n");
	DURAUDIO_WaveClose(pGeode, CHANNEL0_PLAYBACK);

	s->dma_dac.ready=FALSE;
}	

static void nsc_close_record (struct nsc_state *s)
{
	stop_adc(s);

	OS_DbgMsg ("Closing Stream. \n");
	DURAUDIO_WaveClose(pGeode, CHANNEL1_RECORD);

	s->dma_adc.ready=FALSE;
}	

static void nsc_interrupt(int irq, void *dev_id, struct pt_regs *regs);

static int prog_dmabuf(struct nsc_state *s, unsigned rec)
{
	OS_DbgMsg("prog_dmabuf\n");

	if (rec) 
	{
		if( DURAUDIO_WaveOpen (	pGeode, 
								CHANNEL1_RECORD, 
								s->rateadc, 
								s->dma_adc.nChannels, 
								s->dma_adc.BitsPerSample, 
								s->dma_adc.fragsize, 
								INTERVAL_BYTES, 
								s->dma_adc.dmasize,
								s->dma_adc.DMAmapped) )
		{
			set_adc_rate(s, s->rateadc);

			s->dma_adc.ready = TRUE;
		}
		else
		{
			return FALSE;
		}
	}
	else 
	{
		nsc_close_playback(s);

		if( DURAUDIO_WaveOpen (	pGeode, 
								CHANNEL0_PLAYBACK, 
								s->ratedac, 
								s->dma_dac.nChannels, 
								s->dma_dac.BitsPerSample, 
								s->dma_dac.fragsize, 
								INTERVAL_BYTES, 
								s->dma_dac.dmasize,
								s->dma_dac.DMAmapped) )
		{
			OS_DbgMsg("Opening stream.\n");
			set_dac_rate(s, s->ratedac);

			s->dma_dac.ready = TRUE;
			s->dma_dac.count = 0;
		}
		else
		{
			printk("Failed Opening stream (Allocation?)\n");
			return FALSE;
		}
	}

	return TRUE;
}

static void nsc_update_ptr(struct nsc_state *s)
{
	if (s->enable & DAC_RUNNING) 
	{
		long br;
		unsigned size = s->dma_dac.fragsize;
		br = DURAUDIO_UpdateByteCounter  (pGeode, CHANNEL0_PLAYBACK); 
		s->dma_dac.BytesReadyInDMABuffer = br;

		OS_DbgMsg("geodeoss: ISR bytes remaining=%ld\n",br);
		//
		// Output buffer underun protection Logic
		//
		// The test and set are atomic, making this thread-safe
		//
		
		// To avoid starvation for nearly empty buffers
		if ( test_bit(ZF_TRANSFER_SINCE_ZERO,&s->dma_dac.zero_flags)
			 && ( br < ( size + Z_1MS_OF_BYTES ) ) ) {
			// If nsc_write is in the midst of a transfer, don't zero
			// as this will step on the user data
			if ( !test_bit(ZF_TRANSFER_PENDING,&s->dma_dac.zero_flags) ) {
				// Near underflow  We zero the free part of the buffer.
				zero_free_playback_buffer(s);

				// and update the flags
				if (br)
					set_bit(ZF_PARTIAL_ZERO,&s->dma_dac.zero_flags);
				else
					clear_bit(ZF_PARTIAL_ZERO,&s->dma_dac.zero_flags);
				clear_bit(ZF_ZERO_SKIPPED,&s->dma_dac.zero_flags);
			} else
				// Tell nsc_write to zero after the transfer (as one was
				// needed)
				set_bit(ZF_ZERO_SKIPPED,&s->dma_dac.zero_flags);

		} else if ( !br && test_bit(ZF_PARTIAL_ZERO,&s->dma_dac.zero_flags) ) {
			// This logic assures that:
			//   (a) we do zero EVERYTHING
			//   (b) don't zero after we have
			if ( !test_bit(ZF_TRANSFER_PENDING,&s->dma_dac.zero_flags) ) {
				// If nsc_write is in the midst of a transfer, don't zero
				// as this will step on the user data

				// In complete underflow  We zero the buffer (again, probably).
				zero_free_playback_buffer(s);
				clear_bit(ZF_PARTIAL_ZERO,&s->dma_dac.zero_flags);
				clear_bit(ZF_ZERO_SKIPPED,&s->dma_dac.zero_flags);
			} else
				// Tell nsc_write to zero after the transfer (as one was
				// needed)
				set_bit(ZF_ZERO_SKIPPED,&s->dma_dac.zero_flags);

		} else 
			clear_bit(ZF_ZERO_SKIPPED,&s->dma_dac.zero_flags);
			

		if ( br )
		{
			// ZZZ WARN: This logic assumes we catch every interrupt
			s->dma_dac.total_bytes += s->dma_dac.fragsize;

			if (s->dma_dac.DMAmapped)
				s->dma_dac.count += s->dma_dac.fragsize;
			else
				s->dma_dac.count -= s->dma_dac.fragsize;
		}

		wake_up(&s->dma_dac.wait);
	}

	if (s->enable & ADC_RUNNING) 
	{
		s->dma_adc.BytesReadyInDMABuffer = DURAUDIO_UpdateByteCounter  (pGeode, CHANNEL1_RECORD); 

		wake_up(&s->dma_adc.wait);
	}
}

static void nsc_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    struct nsc_state *s;
    struct nsc_card *c = (struct nsc_card *)dev_id;

	unsigned char	IntID;

	IntID = DURAUDIO_InterruptID(pGeode);

	if (IntID==0x04 || IntID==0x08 || IntID==0x0C)
	{
		DURAUDIO_ClearIRQ(pGeode);
		s=&c->channels[0];

		nsc_update_ptr(s);
	}
	else
		return;
}

static void set_mixer(struct nsc_card *card,unsigned int mixer, unsigned int val ) 
{
	unsigned int left,right;

	//
	//  Cleanse input a little 
	//
	right	= ((val >> 8)  & 0xff) ;
	left	= (val  & 0xff) ;

	if(right > 100) 
		right = 100;

	if(left > 100) 
		left = 100;

	card->mix.mixer_state[mixer]=(right << 8) | left;
	card->mix.write_mixer(card,mixer,left,right);
}

static void mixer_push_state(struct nsc_card *card)
{
	int i;

	for(i = 0 ; i < SOUND_MIXER_NRDEVICES ; i++) 
	{
		if( ! supported_mixer(card,i)) continue;

		set_mixer(card,i,card->mix.mixer_state[i]);
	}
}

static int mixer_ioctl(struct nsc_card *card, unsigned int cmd, unsigned long arg)
{
	int i, val=0;

    if (cmd == SOUND_MIXER_INFO) 
	{
		mixer_info info;
		strncpy(info.id, card_names[card->card_type], sizeof(info.id));
		strncpy(info.name,card_names[card->card_type],sizeof(info.name));
		info.modify_counter = card->mix.modcnt;
	
		if (copy_to_user((void *)arg, &info, sizeof(info)))
			return -EFAULT;

		return 0;
	}

	if (cmd == SOUND_OLD_MIXER_INFO) 
	{
		_old_mixer_info info;
		strncpy(info.id, card_names[card->card_type], sizeof(info.id));
		strncpy(info.name,card_names[card->card_type],sizeof(info.name));

		if (copy_to_user((void *)arg, &info, sizeof(info)))
			return -EFAULT;
		return 0;
	}

	if (cmd == OSS_GETVERSION)
		return put_user(SOUND_VERSION, (int *)arg);

	if (_IOC_TYPE(cmd) != 'M' || _IOC_SIZE(cmd) != sizeof(int))
                return -EINVAL;

	if (_IOC_DIR(cmd) == _IOC_READ) 
	{
		switch (_IOC_NR(cmd)) 
		{
			case SOUND_MIXER_RECSRC: // give them the current record source 
			{
				if(!card->mix.recmask_io) 
				{
					val = 0;
				} 
				else 
				{
					val = card->mix.recmask_io(card,1,0);
				}
				break;
			}
		
			case SOUND_MIXER_DEVMASK: // give them the supported mixers 
			{
				val = card->mix.supported_mixers;
				break;
			}

			case SOUND_MIXER_RECMASK: // Arg contains a bit for each supported recording source 
			{
				val = card->mix.record_sources;
				break;
			}

			case SOUND_MIXER_STEREODEVS: // Mixer channels supporting stereo 
			{
				val = card->mix.stereo_mixers;
				break;
			}

			case SOUND_MIXER_CAPS:
			{
				val = SOUND_CAP_EXCL_INPUT;
				break;
			}

			default: // read a specific mixer 
			{
				i = _IOC_NR(cmd);

				if ( ! supported_mixer(card,i)) 
					return -EINVAL;

				//  do we ever want to touch the hardware? 
				//  spin_lock_irqsave(&card->lock, flags);
				//  val = card->mix.read_mixer(card,i);
				//	spin_unlock_irqrestore(&card->lock, flags);
				//

				val = card->mix.mixer_state[i];

				break;
			}
		}

		return put_user(val,(int *)arg);
	}
	
    if (_IOC_DIR(cmd) != (_IOC_WRITE|_IOC_READ))
		return -EINVAL;
	
	card->mix.modcnt++;

	if (get_user(val, (int *)arg))
		return -EFAULT;

	switch (_IOC_NR(cmd)) 
	{
		case SOUND_MIXER_RECSRC: // Arg contains a bit for each recording source 
		{
			if (!card->mix.recmask_io) return -EINVAL;
			if(!val) return 0;
			if(! (val &= card->mix.record_sources)) return -EINVAL;

			card->mix.recmask_io(card,0,val);
			return 0;
		}

		default:
		{
			i = _IOC_NR(cmd);

			if ( ! supported_mixer(card,i)) 
				return -EINVAL;

			set_mixer(card,i,val);

			return 0;
		}
	}
}

static int nsc_open_mixdev(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	struct nsc_card *card = NULL;
	struct pci_dev *pdev;
	struct pci_driver *drvr;

	pci_for_each_dev(pdev) 
	{
		drvr = pci_dev_driver (pdev);

		if (drvr == &geode_pci_driver) 
		{
			card = (struct nsc_card*)pci_get_drvdata (pdev);
			if (!card)
				continue;
			if (card->dev_mixer == minor)
				break;
		}
	}

	if (!card)
		return -ENODEV;

	file->private_data = card;

	return 0;
}

static int nsc_release_mixdev(struct inode *inode, struct file *file)
{
	return 0;
}

static int nsc_ioctl_mixdev(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct nsc_card *card = (struct nsc_card *)file->private_data;

	return mixer_ioctl(card, cmd, arg);
}

static /*const*/ struct file_operations nsc_mixer_fops = 
{
	owner:		THIS_MODULE,
	llseek:         no_llseek,
	ioctl:          nsc_ioctl_mixdev,
	open:           nsc_open_mixdev,
	release:        nsc_release_mixdev,
};

//
//  In this loop, dma_adc.count signifies the amount of data thats waiting
//	to be copied to the user's buffer.  it is filled by the interrupt
//	handler and drained by this loop. 
//
static ssize_t nsc_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	struct nsc_state *s = (struct nsc_state *)file->private_data;
	DECLARE_WAITQUEUE(wait, current);
	ssize_t			ret;
	unsigned long   BytesTransferred, swPointer, BytesReadyToRead;

	OS_DbgMsg("nsc_read - count: %d\n", count);

	if (ppos != &file->f_pos)
		return -ESPIPE;

	if(s->dma_adc.DMAmapped)
		return -ENXIO;

	if (!access_ok(VERIFY_WRITE, buffer, count))
		return -EFAULT;

	if (!s->dma_adc.ready)
	{
		OS_DbgMsg ("nsc_read: ");
		if (!prog_dmabuf(s, 1))
		{
			OS_DbgMsg("prog_dmabuf failed!\n");
			return -ENOMEM;
		}
	}

	//
	//   Start ADC
	//
	if(!(s->enable & ADC_RUNNING))
	{
		start_adc(s);	
	}

	if (!count)
		return count;

	ret = count;

	add_wait_queue(&s->dma_adc.wait, &wait);
	swPointer = 0;
	//
	//   Just wait if we still don't have data to fill up the user block
	//
	while (count)
	{
		if (file->f_flags & O_NONBLOCK) 
		{
			ret = ret ? ret : -EAGAIN;
			OS_DbgMsg("Un-Blocked!\n");
			goto out1; 
		}

		__set_current_state(TASK_INTERRUPTIBLE);
		schedule();

		if (signal_pending(current)) 
		{
			printk("Signal Pending!\n");
			ret = ret ? ret : -ERESTARTSYS;
			goto out1;
		}
		
		BytesReadyToRead = DURAUDIO_UpdateByteCounter (pGeode, CHANNEL1_RECORD); 

		if (BytesReadyToRead > count)
			BytesReadyToRead = count;

		OS_DbgMsg("BytesReadyToRead: %d,  count: %d\n", BytesReadyToRead, count);

		BytesTransferred = DURAUDIO_TransferAudioData (pGeode, CHANNEL1_RECORD, (unsigned char *) (buffer + swPointer), BytesReadyToRead); 

		count		-= BytesReadyToRead;
		swPointer	+= BytesReadyToRead;
	} 

out1:
	remove_wait_queue(&s->dma_adc.wait, &wait);
	set_current_state(TASK_RUNNING);

	return ret;
}

static ssize_t nsc_write(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
	struct nsc_state *s = (struct nsc_state *)file->private_data;
	ssize_t			ret;
	size_t			cnt;

	unsigned long   BytesTransferred, swPointer, BytesAvailableToFill, BytesInDMABuffer;
	
	OS_DbgMsg("nsc_write - count: %d\n", count);

	if (ppos != &file->f_pos)
		return -ESPIPE;

	if(s->dma_dac.DMAmapped)
		return -ENXIO;

	if (!access_ok(VERIFY_READ, buffer, count))
		return -EFAULT;

	if (!s->dma_dac.ready)
	{
		OS_DbgMsg ("nsc_write: ");
		if (!prog_dmabuf(s, 0))
		{
			OS_DbgMsg("prog_dmabuf failed!\n");
			return -ENOMEM;
		} 
	}

	if (!count)
		return count;

	ret = count;

	swPointer = 0;
	cnt = count;

	//
	//  We try to keep the DMA buffer full.
	//	If we don't have any room to put the data, we wait and try again.
	//
	while (count)
	{
		BytesInDMABuffer = get_playback_bytes_used(s);

		BytesAvailableToFill = s->dma_dac.dmasize - BytesInDMABuffer;
	
		OS_DbgMsg("BytesInDMABuffer: %d  BytesAvailableToFill: %d\n", BytesInDMABuffer, BytesAvailableToFill);

		if (BytesAvailableToFill > count)
			BytesAvailableToFill = count;

		BytesTransferred = 0;
		if (BytesAvailableToFill)
		{
			// The bit prevents the ISR from doing a underrun "zero the buffer"
			set_bit(ZF_TRANSFER_PENDING,&s->dma_dac.zero_flags);

			BytesTransferred  = DURAUDIO_TransferAudioData (pGeode,
				CHANNEL0_PLAYBACK, (unsigned char *) (buffer + swPointer),
				BytesAvailableToFill); 

			// Transfer since tells the ISR to zero on next underrun
			set_bit(ZF_TRANSFER_SINCE_ZERO,&s->dma_dac.zero_flags);

			// If an underrun event happened during the Transfer
			// we need to handle the buffer zeroing here.
			//  -- if the buffer is full, then nothing will happen, but
			// this is safe in any case
			if ( test_bit(ZF_ZERO_SKIPPED,&s->dma_dac.zero_flags) ) {
				zero_free_playback_buffer(s);
				clear_bit(ZF_ZERO_SKIPPED,&s->dma_dac.zero_flags);
			}
			clear_bit(ZF_TRANSFER_PENDING,&s->dma_dac.zero_flags);

			// Update the counters and pointers based on the data
			// actually transferred (the returned value)
			count			 -= BytesTransferred;
			swPointer		 += BytesTransferred;
			s->dma_dac.count += BytesTransferred;

			// ZZZ RACE, even though the ISR updates this value we still
			// have to add the recent bytes.  The symptom of the race
			// is the counters may get out of sync.  Causing an incomplete
			// "TransferAudio" above.
			s->dma_dac.BytesReadyInDMABuffer += BytesTransferred;

			//
			//   Is the DAC running yet?
			//
			if(!(s->enable & DAC_RUNNING))
			{
				OS_DbgMsg("Trying to start DAC...\n");

				start_dac(s,START_DAC_DEFAULT);	
			}
		}

		// Is it time to sleep?
		// Either condition indicates a full buffer.
		if ( !BytesAvailableToFill || 
				(BytesTransferred != BytesAvailableToFill ) ) {
			if (file->f_flags & O_NONBLOCK) 
			{
				ret = ret ? ret : -EAGAIN;
				OS_DbgMsg("Un-Blocked!\n");
				goto out1; 
			}

			interruptible_sleep_on(&s->dma_dac.wait);
			if (signal_pending(current)) 
			{
				OS_DbgMsg("Signal Pending!\n");
				ret = ret ? ret : -ERESTARTSYS;
				goto out1;
			}
		}
	}

out1:

	return ret;
}

//
// No kernel lock - we have our own spinlock 
//
static unsigned int nsc_poll(struct file *file, struct poll_table_struct *wait)
{
	struct nsc_state *s = (struct nsc_state *)file->private_data;
	unsigned int mask = 0;

	OS_DbgMsg("*** nsc_poll *** \n");

	if (file->f_mode & FMODE_WRITE) 
	{
		if (!s->dma_dac.ready && (!prog_dmabuf(s, 0))) 
			return 0;
	}

	if (file->f_mode & FMODE_READ) 
	{
	  	if (!s->dma_adc.ready && (!prog_dmabuf(s, 1)))
			return 0;
	}

	if (file->f_mode & FMODE_WRITE)
		poll_wait(file, &s->dma_dac.wait, wait);

	if (file->f_mode & FMODE_READ)
		poll_wait(file, &s->dma_adc.wait, wait);

	if (file->f_mode & FMODE_READ) 
	{
		if (s->dma_adc.count >= (signed)s->dma_adc.fragsize)
			mask |= POLLIN | POLLRDNORM;
	}

	if (file->f_mode & FMODE_WRITE) 
	{
		if (s->dma_dac.DMAmapped) 
		{
			if (s->dma_dac.count >= (signed)s->dma_dac.fragsize) 
				mask |= POLLOUT | POLLWRNORM;
		}
		else 
		{
			OS_DbgMsg("POLL **** count: %d\n", s->dma_dac.count);

			//if ((signed)s->dma_dac.dmasize >= (signed) s->dma_dac.count + (signed)s->dma_dac.fragsize)
			{
				mask |= POLLOUT | POLLWRNORM;
			}
		}
	}

	return mask;
}

static int nsc_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct nsc_state *s = (struct nsc_state *)file->private_data;
	struct dmabuf *db;
	int ret = 0;
	unsigned long size, Channel;

	OS_DbgMsg("nsc_mmap:\n");

	lock_kernel();
	
	if (vma->vm_flags & VM_WRITE) 
	{
		Channel=CHANNEL0_PLAYBACK;

		s->dma_dac.DMAmapped = TRUE;

		if (prog_dmabuf(s, 0)) 
		{
			db = &s->dma_dac;
		}
		else
		{
			s->dma_dac.DMAmapped = FALSE;
			ret = -EAGAIN;
			goto out;
		}
	} 
	else 
	{
		if (vma->vm_flags & VM_READ) 
		{
			Channel=CHANNEL1_RECORD;

			s->dma_adc.DMAmapped = TRUE;

			if (prog_dmabuf(s, 1)) 
				db = &s->dma_adc;
			else
			{
				s->dma_adc.DMAmapped = FALSE;
				ret = -EAGAIN;
				goto out;
			}
		} 
		else 
		{
			ret = -EINVAL;
			goto out;
		}
	}

	if (vma->vm_pgoff != 0) 
	{
		ret = -EINVAL;
		goto out;
	}
	
	size = vma->vm_end - vma->vm_start;

	OS_DbgMsg("size: %d   Allocated DMA size: %d\n", size, pGeode->AudioChannel[Channel].DMA_AllocInfo.Size);

	if (size > pGeode->AudioChannel[Channel].DMA_AllocInfo.Size) 
	{
		db->DMAmapped = FALSE;
		ret = -EINVAL;
		goto out;
	}

	//
	//   Save the mapping structure
	//
	memcpy(&db->vma, vma, sizeof(struct vm_area_struct));

	db->rawbuf = (void *) pGeode->AudioChannel[Channel].DMA_AllocInfo.VirtualAddress;

	if (remap_page_range(vma->vm_start, virt_to_phys(db->rawbuf), size, vma->vm_page_prot)) 
	{
		db->DMAmapped = FALSE;
		ret = -EAGAIN;
		goto out;
	}

	db->DMAmapped	= TRUE;
	db->ready		= TRUE;

out:
	unlock_kernel();
	return ret;
}

int drain_dac (struct nsc_state *s)
{
	unsigned long userBytes;

	userBytes = get_playback_bytes_used(s);
	if( !userBytes )
		return 0; // We're done

	start_dac(s,START_DAC_START_ANY);	
	if ( ! (s->enable & DAC_RUNNING ) ) {
		printk(KERN_WARNING "geodeoss: drain_dac: couldn't start DAC\n");
		return -EFAULT;
	}
			
	//
	//	Play the remaining buffered bytes 
	//
	do {
		interruptible_sleep_on(&s->dma_dac.wait);

		if (signal_pending(current)) {
			printk("drain_dac: Signal Pending!\n");
			userBytes = 0; // abandon the drain on signal, but don't return ERR
		} else
			userBytes = get_playback_bytes_used(s);

	} while (userBytes);

	return 0;
}

static int nsc_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct nsc_state	*s = (struct nsc_state *)file->private_data;
	int					val=0;
    audio_buf_info		abinfo;
    count_info			cinfo;

	OS_DbgMsg("geodeoss: nsc_ioctl: cmd %08X\n", cmd);
	
	switch (cmd) 
	{
		case OSS_GETVERSION:
			return put_user(SOUND_VERSION, (int *)arg);

		case SNDCTL_DSP_SYNC:
		{
			OS_DbgMsg("SNDCTL_DSP_SYNC\n");

			//
			//   The app wants us to flush all the data we have in our buffers (If we are playing)
			//

			if ((file->f_mode & FMODE_WRITE) && (s->enable & DAC_RUNNING))
				return drain_dac(s);

			return 0;
		}

		case SNDCTL_DSP_SETDUPLEX:
			OS_DbgMsg("SNDCTL_DSP_SETDUPLEX\n");
			// XXX fix 
			return 0;

		case SNDCTL_DSP_GETCAPS:
		{
			OS_DbgMsg("SNDCTL_DSP_GETCAPS\n");
			return put_user(DSP_CAP_DUPLEX | DSP_CAP_REALTIME | DSP_CAP_TRIGGER | DSP_CAP_MMAP, (int *)arg);
		}

		case SNDCTL_DSP_RESET:
		{
			OS_DbgMsg("SNDCTL_DSP_RESET\n");

			if (file->f_mode & FMODE_WRITE) 
			{
				stop_dac(s);
				synchronize_irq();
				s->dma_dac.ready = FALSE;			
			}

			if (file->f_mode & FMODE_READ) 
			{
				stop_adc(s);
				synchronize_irq();
				s->dma_dac.ready = FALSE;			
			}
			return 0;
		}

        case SNDCTL_DSP_SPEED:
		{
			OS_DbgMsg("SNDCTL_DSP_SPEED\n");

			if (get_user(val, (int *)arg))
				return -EFAULT;

			if (val >= 0) 
			{
				if (file->f_mode & FMODE_READ) 
				{
					stop_adc(s);
					set_adc_rate(s, val);
				}

				if (file->f_mode & FMODE_WRITE) 
				{
					stop_dac(s);
					set_dac_rate(s, val);
				}
			}

			OS_DbgMsg("Rate: %d\n", (file->f_mode & FMODE_READ) ? s->rateadc : s->ratedac);

			return put_user((file->f_mode & FMODE_READ) ? s->rateadc : s->ratedac, (int *)arg);
		}

        case SNDCTL_DSP_STEREO:
		{
			OS_DbgMsg("SNDCTL_DSP_STEREO\n");

			if (get_user(val, (int *)arg))
				return -EFAULT;

			if (file->f_mode & FMODE_READ) 
			{
				if (val)
					s->dma_adc.nChannels=2;
				else
				{
					s->dma_adc.nChannels=1;
				}

			}

			if (file->f_mode & FMODE_WRITE) 
			{
				if (val)
					s->dma_dac.nChannels=2;
				else
					s->dma_dac.nChannels=1;
			}

			return put_user(val, (int *)arg);
		}

        case SNDCTL_DSP_CHANNELS:
		{
			OS_DbgMsg("SNDCTL_DSP_CHANNELS\n");

			if (get_user(val, (int *)arg))
				return -EFAULT;

			if (file->f_mode & FMODE_READ) 
			{
				s->dma_adc.nChannels=val;
			}

			if (file->f_mode & FMODE_WRITE) 
			{
				s->dma_dac.nChannels=val;
			}

			return put_user(val, (int *)arg);
		}

		case SNDCTL_DSP_GETFMTS: // Returns a mask 
		{
			OS_DbgMsg("SNDCTL_DSP_GETFMTS\n");
			return put_user(AFMT_U8|AFMT_S16_LE, (int *)arg);
		}

		case SNDCTL_DSP_SETFMT: // Selects ONE fmt
		{
			OS_DbgMsg("SNDCTL_DSP_SETFMT\n");

			if (get_user(val, (int *)arg))
				return -EFAULT;

			if (val != AFMT_QUERY) 
			{
				if (file->f_mode & FMODE_READ) 
				{
					if (val == AFMT_S16_LE)
					{
						s->dma_adc.BitsPerSample=16;
					}
					else
					{
						//
						//  Our hardware only does 16bit Stereo on DirectDMA.  
						//
						if (s->dma_adc.DMAmapped)
						{
							printk("We cannot do DirectDMA in anything different than 16bit Signed Little Endian format.\n");
							return -EFAULT;
						}

						s->dma_adc.BitsPerSample=8;

						if (val != AFMT_U8)
						{
							//
							//	This format is not supported
							//
							printk ("We don't support this format. Only Unsigned 8bit or 16bit Signed Little Endian formats.\n");
							return -EFAULT;
						}
					}  
				}
				
				if (file->f_mode & FMODE_WRITE) 
				{
					OS_DbgMsg("Format: %08X\n", val);
					if (val == AFMT_S16_LE)
					{
						s->dma_dac.BitsPerSample=16;
					}
					else
					{
						//
						//  Our hardware only does 16bit Stereo on DirectDMA.  
						//
						if (s->dma_dac.DMAmapped)
						{
							printk("We cannot do DirectDMA in anything different than 16bit Signed Little Endian format.\n");
							return -EFAULT;
						}

						s->dma_dac.BitsPerSample=8;

						if (val != AFMT_U8)
						{
							//
							//	This format is not supported
							//
							printk("We don't support this format. Only Unsigned 8bit or 16bit Signed Little Endian formats.\n");
							return -EFAULT;
						}
					}	
				}
			}

			OS_DbgMsg("BitsPerSample: %d  val:%d\n",s->dma_dac.BitsPerSample, val );

			return put_user(val, (int *)arg);
		}		

		case SNDCTL_DSP_POST:
		{
			OS_DbgMsg("SNDCTL_DSP_POST\n");
			return 0;
		}

        case SNDCTL_DSP_GETTRIGGER:
		{
			OS_DbgMsg("SNDCTL_DSP_GETTRIGGER\n");

			val = 0;

			if ((file->f_mode & FMODE_READ) && (s->enable & ADC_RUNNING))
				val |= PCM_ENABLE_INPUT;

			if ((file->f_mode & FMODE_WRITE) && (s->enable & DAC_RUNNING)) 
				val |= PCM_ENABLE_OUTPUT;

			return put_user(val, (int *)arg);
		}

		case SNDCTL_DSP_SETTRIGGER:
		{
			OS_DbgMsg("SNDCTL_DSP_SETTRIGGER\n");

			if (get_user(val, (int *)arg))
				return -EFAULT;

			if (file->f_mode & FMODE_READ) 
			{
				if (val & PCM_ENABLE_INPUT) 
				{
					OS_DbgMsg("SNDCTL_DSP_SETTRIGGER: FMODE_READ - START\n");
					OS_DbgMsg("s->enable: %d\n", s->enable);

					if (s->enable & ADC_RUNNING) 
						return TRUE;

					if(s->dma_adc.ready)
						start_adc(s);
				} 
				else
				{
					stop_adc(s);
				}
			}

			if (file->f_mode & FMODE_WRITE) 
			{
				if (val & PCM_ENABLE_OUTPUT) 
				{
					OS_DbgMsg("SNDCTL_DSP_SETTRIGGER: FMODE_WRITE - START\n");
					OS_DbgMsg("s->enable: %d\n",		s->enable);
					OS_DbgMsg("s->dma_dac.ready: %s\n", s->dma_dac.ready?"TRUE":"FALSE");

					if (s->enable & DAC_RUNNING) 
						return TRUE;

					if(s->dma_dac.ready)
						start_dac(s,START_DAC_START_ANY);
				} 
				else
				{
					stop_dac(s);
				}
			}
			return 0;
		}

		case SNDCTL_DSP_GETOSPACE:
		{
			long	BytesInBuffer;

			OS_DbgMsg("SNDCTL_DSP_GETOSPACE: \n");

			if (!(file->f_mode & FMODE_WRITE))
				return -EINVAL;

			abinfo.fragsize		= s->dma_dac.fragsize;
			abinfo.fragstotal	= s->dma_dac.dmasize / s->dma_dac.fragsize;

			if (!s->dma_dac.ready)
			{
				abinfo.bytes		= (s->dma_dac.dmasize);
				abinfo.fragments	= abinfo.bytes / s->dma_dac.fragsize;    

				OS_DbgMsg("fragsize: %d fragstotal:%d   What can be read without blocking: Bytes: %d  fragments: %d\n", 
																							abinfo.fragsize, 
																							abinfo.fragstotal, 
																							abinfo.bytes, 
																							abinfo.fragments);
				
				return copy_to_user((void *)arg, &abinfo, sizeof(abinfo)) ? -EFAULT : 0;
			}

			if (s->enable & DAC_RUNNING) 
				BytesInBuffer = DURAUDIO_UpdateByteCounter  (pGeode, CHANNEL0_PLAYBACK); 
			else
				BytesInBuffer = 0;

			abinfo.bytes		= (s->dma_dac.dmasize) - BytesInBuffer;
			abinfo.fragments	= abinfo.bytes / s->dma_dac.fragsize;    
			
			OS_DbgMsg("Space: fragsize: %d bytes: %d fragstotal:%d fragments: %d\n", s->dma_dac.fragsize, abinfo.bytes, abinfo.fragstotal, abinfo.fragments);

			return copy_to_user((void *)arg, &abinfo, sizeof(abinfo)) ? -EFAULT : 0;
		}

		case SNDCTL_DSP_GETISPACE:
		{
			OS_DbgMsg("SNDCTL_DSP_GETISPACE\n");

			if (!(file->f_mode & FMODE_READ))
				return -EINVAL;

			if (!s->dma_adc.ready)
				return -EINVAL;

			abinfo.fragsize		= s->dma_adc.fragsize;
			abinfo.bytes		= s->dma_dac.BytesReadyInDMABuffer;

			abinfo.fragstotal	= s->dma_adc.dmasize / s->dma_adc.fragsize;
			abinfo.fragments	= abinfo.bytes / s->dma_adc.fragsize;   
			
			OS_DbgMsg("fragsize: %d bytes: %d fragstotal:%d fragments: %d\n", s->dma_adc.fragsize, abinfo.bytes, abinfo.fragstotal, abinfo.fragments);

			return copy_to_user((void *)arg, &abinfo, sizeof(abinfo)) ? -EFAULT : 0;
		}

		case SNDCTL_DSP_NONBLOCK:
		{
			OS_DbgMsg("SNDCTL_DSP_NONBLOCK\n");

			file->f_flags |= O_NONBLOCK;
			return 0;
		}

        case SNDCTL_DSP_GETODELAY:
		{
			OS_DbgMsg("SNDCTL_DSP_GETODELAY\n");

			if (!(file->f_mode & FMODE_WRITE))
				return -EINVAL;

			if (!s->dma_dac.ready) 
				return 1;
			
			val = (int) DURAUDIO_UpdateByteCounter  (pGeode, CHANNEL0_PLAYBACK);

			if ((s->enable & DAC_RUNNING) && (val <= 0))
			{
				nsc_close_playback(s);
				return 1;
			}

			OS_DbgMsg("Bytes Ready in DMA buffer: %d\n", val);

			return put_user(val, (int *) arg);
		}

        case SNDCTL_DSP_GETIPTR:
		{
			OS_DbgMsg("SNDCTL_DSP_GETIPTR\n");

			if (!(file->f_mode & FMODE_READ))
				return -EINVAL;

			if (!s->dma_adc.ready)
				return -EINVAL;

			cinfo.bytes		= s->dma_adc.total_bytes;
			cinfo.blocks	= s->dma_adc.count;
			cinfo.ptr		= DURAUDIO_GetPosition(pGeode, CHANNEL1_RECORD);
			cinfo.ptr		= cinfo.ptr / ((16 / s->dma_adc.BitsPerSample) * (2 / s->dma_adc.nChannels));

			if (s->dma_adc.DMAmapped)
				s->dma_dac.count = 0;

			return copy_to_user((void *)arg, &cinfo, sizeof(cinfo));
		}

        case SNDCTL_DSP_GETOPTR:
		{
			OS_DbgMsg("SNDCTL_DSP_GETOPTR\n");

			if (!(file->f_mode & FMODE_WRITE))
				return -EINVAL;

			if (!s->dma_dac.ready)
				return -EINVAL;

			cinfo.bytes		= s->dma_dac.total_bytes;
			cinfo.blocks	= s->dma_dac.count;
			cinfo.ptr		= DURAUDIO_GetPosition(pGeode, CHANNEL0_PLAYBACK);
			cinfo.ptr		= cinfo.ptr / ((16 / s->dma_dac.BitsPerSample) * (2 / s->dma_dac.nChannels));

			if (s->dma_dac.DMAmapped)
				s->dma_dac.count = 0;

			return copy_to_user((void *)arg, &cinfo, sizeof(cinfo));
		}

        case SNDCTL_DSP_GETBLKSIZE:
		{
			OS_DbgMsg("SNDCTL_DSP_GETBLKSIZE\n");

			if (file->f_mode & FMODE_WRITE) 
			{
				OS_DbgMsg("DAC fragsize: %d  dmasize: %d*****************************************\n", s->dma_dac.fragsize, s->dma_dac.dmasize);
				return put_user(s->dma_dac.fragsize, (int *)arg);
			}
			else
			{
				OS_DbgMsg("ADC fragsize: %d  dmasize: %d*****************************************\n", s->dma_adc.fragsize, s->dma_adc.dmasize);
				return put_user(s->dma_adc.fragsize, (int *)arg);
			}
		}

        case SNDCTL_DSP_SETFRAGMENT:
		{
			if (get_user(val, (int *)arg))
				return -EFAULT;

			OS_DbgMsg("geodeoss: SETFRAGMENT: %0x\n",val);

			if (file->f_mode & FMODE_READ) 
			{
				s->dma_adc.ossfragshift	= val & 0xffff;
				s->dma_adc.ossmaxfrags	= (val >> 16) & 0xffff;

				s->dma_adc.fragsize		= 1;
				s->dma_adc.fragsize		<<= (val & 0x0000FFFF);
				s->dma_adc.ossfragsize	= s->dma_adc.fragsize;

				//
				//   0x7FFF means no limits for number of frags.  Set it to DEFAULT_NUMBER_OF_FRAGMENTS.
				//
				if(s->dma_adc.ossmaxfrags==0x7FFF)
					s->dma_adc.ossmaxfrags = DEFAULT_NUMBER_OF_FRAGMENTS;				
				
				OS_DbgMsg("Requested: Fragsize: %d  MaxFrags: %d*****************************************\n", s->dma_adc.ossfragsize, s->dma_adc.ossmaxfrags);

				s->dma_adc.dmasize  = s->dma_adc.fragsize * s->dma_adc.ossmaxfrags;

				if (s->dma_adc.dmasize > (DEFAULT_DMA_BUFFER_SIZE))
				{
					s->dma_adc.dmasize  = DEFAULT_DMA_BUFFER_SIZE;
					s->dma_adc.fragsize = DEFAULT_FRAGMENT_SIZE;

					OS_DbgMsg("Adjusted to:  Fragsize: %d  MaxFrags: %d DMAsize: %d *****************************************\n", s->dma_adc.ossfragsize, (DEFAULT_DMA_BUFFER_SIZE / s->dma_adc.ossfragsize), s->dma_adc.dmasize);
				}

				s->dma_adc.ready=FALSE;
			}

			if (file->f_mode & FMODE_WRITE) 
			{
				s->dma_dac.ossfragshift = val & 0xffff;
				s->dma_dac.ossmaxfrags = (val >> 16) & 0xffff;

				s->dma_dac.ossfragsize = 1;
				s->dma_dac.ossfragsize <<= (val&0x0000FFFF);

				//
				//   0x7FFF means no limits for number of frags.  Set it to DEFAULT_NUMBER_OF_FRAGMENTS.
				//
				if(s->dma_dac.ossmaxfrags==0x7FFF)
					s->dma_dac.ossmaxfrags = DEFAULT_NUMBER_OF_FRAGMENTS;

				OS_DbgMsg("Requested: Fragsize: %d  MaxFrags: %d*****************************************\n", s->dma_dac.ossfragsize, s->dma_dac.ossmaxfrags);

				s->dma_dac.dmasize = s->dma_dac.ossfragsize * s->dma_dac.ossmaxfrags;

				if (s->dma_dac.dmasize > (DEFAULT_DMA_BUFFER_SIZE))
				{
					s->dma_dac.dmasize  = DEFAULT_DMA_BUFFER_SIZE;
					s->dma_dac.fragsize = DEFAULT_FRAGMENT_SIZE;

					OS_DbgMsg("Adjusted to:  Fragsize: %d  MaxFrags: %d DMAsize: %d *****************************************\n", s->dma_dac.ossfragsize, (DEFAULT_DMA_BUFFER_SIZE / s->dma_dac.ossfragsize), s->dma_dac.dmasize);
				}

				s->dma_dac.fragsize	= s->dma_dac.ossfragsize;

				s->dma_dac.ready=FALSE;
			}

			return 0;
		}

        case SNDCTL_DSP_SUBDIVIDE:
		{
			OS_DbgMsg("geodeoss: SNDCTL_DSP_SUBDIVIDE\n");

			if ((file->f_mode & FMODE_READ && s->dma_adc.subdivision) || (file->f_mode & FMODE_WRITE && s->dma_dac.subdivision))
				return -EINVAL;

			if (get_user(val, (int *)arg))
				return -EFAULT;

			if (val != 1 && val != 2 && val != 4)
				return -EINVAL;

			if (file->f_mode & FMODE_READ)
				s->dma_adc.subdivision = val;

			if (file->f_mode & FMODE_WRITE)
				s->dma_dac.subdivision = val;

			return 0;
		}

        case SOUND_PCM_READ_RATE:
			OS_DbgMsg("geodeoss: SOUND_PCM_READ_RATE\n");
			return put_user((file->f_mode & FMODE_READ) ? s->rateadc : s->ratedac, (int *)arg);

        case SOUND_PCM_READ_CHANNELS:
			OS_DbgMsg("geodeoss: SOUND_PCM_READ_CHANNELS\n");
			return put_user((s->fmt & ((file->f_mode & FMODE_READ) ? (NSC_FMT_STEREO << NSC_ADC_SHIFT) 
						   : (NSC_FMT_STEREO << NSC_DAC_SHIFT))) ? 2 : 1, (int *)arg);

        case SOUND_PCM_READ_BITS:
			OS_DbgMsg("geodeoss: SOUND_PCM_READ_BITS\n");
			return put_user((s->fmt & ((file->f_mode & FMODE_READ) ? (NSC_FMT_16BIT << NSC_ADC_SHIFT) 
						   : (NSC_FMT_16BIT << NSC_DAC_SHIFT))) ? 16 : 8, (int *)arg);

        case SOUND_PCM_WRITE_FILTER:
			OS_DbgMsg("geodeoss: SOUND_PCM_WRITE_FILTER\n");
        case SNDCTL_DSP_SETSYNCRO:
			OS_DbgMsg("geodeoss: SNDCTL_DSP_SETSYNCRO\n");
        case SOUND_PCM_READ_FILTER:
			OS_DbgMsg("geodeoss: SOUND_PCM_READ_FILTER\n");
                return -EINVAL;
		
	}
	return -EINVAL;
}

static int nsc_open(struct inode *inode, struct file *file)
{
	struct nsc_card *c = devs;
	struct nsc_state *s = NULL;

	OS_DbgMsg("geodeoss: nsc_open\n");

	s=&c->channels[0];

	if (!s)
		return -ENODEV;

	file->private_data = s;

	//
	// wait for device to become free 
	//
	down(&s->open_sem);

	while (s->open_mode & file->f_mode) 
	{
		if (file->f_flags & O_NONBLOCK) 
		{
			up(&s->open_sem);
			return -EWOULDBLOCK;
		}

		up(&s->open_sem);
		interruptible_sleep_on(&s->open_wait);

		if (signal_pending(current))
			return -ERESTARTSYS;

		down(&s->open_sem);
	}


	//
	//	Initializing everything...
	//
	if (file->f_mode & FMODE_READ) 
	{
		s->dma_adc.ossfragsize				= s->dma_adc.fragsize = DEFAULT_FRAGMENT_SIZE;		
		s->dma_adc.dmasize					= DEFAULT_DMA_BUFFER_SIZE;
											
		s->dma_adc.DMAmapped				= FALSE;
		s->dma_adc.total_bytes				= 0;		
		s->dma_adc.rawbuf					= NULL;
		s->dma_adc.ready					= FALSE;	
		s->dma_adc.BytesReadyInDMABuffer	= 0;

		//
		//  Our hardware only does 16bit Stereo on DirectDMA.  
		//	Assume it as default.
		//
		s->dma_adc.BitsPerSample=16;
		s->dma_adc.nChannels=2;

		set_adc_rate(s, 8000);
	}

	if (file->f_mode & FMODE_WRITE) 
	{
		s->dma_dac.ossfragsize				= s->dma_dac.fragsize = DEFAULT_FRAGMENT_SIZE;		
		s->dma_dac.dmasize					= DEFAULT_DMA_BUFFER_SIZE;
											
		s->dma_dac.DMAmapped				= FALSE;
		s->dma_dac.total_bytes				= 0;		
		s->dma_dac.rawbuf					= NULL;
		s->dma_dac.ready					= FALSE;	
		s->dma_dac.BytesReadyInDMABuffer	= 0;

		//
		//  Our hardware only does 16bit Stereo on DirectDMA.  
		//	Assume it as default.
		//
		s->dma_dac.BitsPerSample=16;
		s->dma_dac.nChannels=2; 

		set_dac_rate(s, 8000);
	}

	s->open_mode |= file->f_mode & (FMODE_READ | FMODE_WRITE);

	up(&s->open_sem);

	return 0;
}

static int nsc_release(struct inode *inode, struct file *file)
{
	struct nsc_state	*s = (struct nsc_state *)file->private_data;

	OS_DbgMsg("nsc_release\n");

	if ((file->f_mode & FMODE_WRITE))
	{
		if(s->dma_dac.ready) 
		{
			//
			//	Flushing DAC:
			//
			//  Do we have bytes in the buffer and the DAC is ready but not running yet?
			//

			OS_DbgMsg("Flushing DAC\n");
			drain_dac (s);

		}
	
		nsc_close_playback(s);
	}

	if (file->f_mode & FMODE_READ) 
	{
		nsc_close_record(s);
	}

	lock_kernel();

	down(&s->open_sem);
		
	s->open_mode &= (~file->f_mode) & (FMODE_READ|FMODE_WRITE);

	s->dma_dac.rawbuf	 = s->dma_adc.rawbuf	= NULL;
	s->dma_dac.DMAmapped = s->dma_adc.DMAmapped = FALSE;
	s->dma_dac.ready	 = s->dma_adc.ready		= FALSE;

	up(&s->open_sem);
	wake_up(&s->open_wait);
	unlock_kernel();

	return 0;
}

static struct file_operations nsc_audio_fops = 
{
	owner:			THIS_MODULE,
	llseek:         no_llseek,
	read:           nsc_read,
	write:          nsc_write,
	poll:           nsc_poll,
	ioctl:          nsc_ioctl,
	mmap:           nsc_mmap,
	open:           nsc_open,
	release:        nsc_release,
};

//
//  This function will find out in the Bios/Hardware if the device 
//  has power management capabilities and return 1 if success.
//
int parse_power(struct nsc_card *card, struct pci_dev *pcidev)
{
	return 1;
}

static int __init geode_probe(struct pci_dev *pcidev,const struct pci_device_id *pdid)
{
	int		card_type = pdid->driver_data;
	u32		n;
	int		iobase;
	int		ret;

	struct nsc_card		*card;
	struct nsc_state	*nsc;
	struct pm_dev		*pmdev;
	struct nsc_state	*s;

	//
	// When built into the kernel, we only print version if device is found 
	//

	OS_DbgMsg(KERN_WARNING "geodeoss: Probing hardware\n");

#ifndef MODULE
	static int printed_version = 0;

	if (!printed_version++) {
		printk(version);
		printk(KERN_INFO "geodeoss: use_pm=%d\n",use_pm);
	}
#endif

	OS_DbgMsg( "Found Card: %s\n", card_names[card_type]);

	if ((ret=pci_enable_device(pcidev)))
	{
		OS_DbgMsg(KERN_WARNING "geodeoss: FAILED - pci_enable_device!\n");
		return ret;
	}
			
	iobase = pci_resource_start(pcidev,0);

	if (!iobase) //  || !(pci_resource_flags(pcidev, 0 ) & IORESOURCE_IO))
	{
		OS_DbgMsg(KERN_WARNING "geodeoss: FAILED - No IO Base!\n");
		return -ENODEV;
	}

	if(pcidev->irq == 0)
	{
		pcidev->irq = 5;
		OS_DbgMsg(KERN_WARNING "geodeoss: No IRQ!  Assuming IRQ %d\n", pcidev->irq);
	}

	pGeode=DURAUDIO_Initialize((unsigned int) pcidev->irq);

	if (!pGeode)
		return -ENODEV;


	if (card_type==TYPE_GX2)
	{
		//
		// stake our claim on the iospace 
		//
		printk("Claiming IO space.\n");
		if( request_region(iobase, 128, card_names[card_type]) == NULL )
		{
			OS_DbgMsg(KERN_WARNING "geodeoss: can't allocate 128 bytes I/O at 0x%4.4x\n", iobase);
			return -EBUSY;
		}
		else
			gAllocatedIOResources = TRUE;
	}

	//
	// Just to be sure 
	//
	pci_set_master(pcidev);

	card = kmalloc(sizeof(struct nsc_card), GFP_KERNEL);

	if(card == NULL)
	{
		OS_DbgMsg(KERN_WARNING "geodeoss: out of memory\n");

		if (card_type==TYPE_GX2)
		{
			release_region(iobase, 128);
			return -ENOMEM;
		}
	}
	
	memset(card, 0, sizeof(*card));
	card->pcidev = pcidev;


	card->iobase	= iobase;
	card->card_type	= card_type;
	card->irq		= pcidev->irq;

	spin_lock_init		(&card->lock);
	init_waitqueue_head	(&card->suspend_queue);

	card->dock_mute_vol = 50;

	devs=card;

	s = &card->channels[0];

	s->card = card;

	init_waitqueue_head(&s->dma_adc.wait);
	init_waitqueue_head(&s->dma_dac.wait);
	init_waitqueue_head(&s->open_wait);

	init_MUTEX(&s->open_sem);

	if(s->dma_adc.ready || s->dma_dac.ready || s->dma_adc.rawbuf)
		OS_DbgMsg("geodeoss: BOTCH!\n");

	s->enable=0;

	//
	// register devices 
	//
	s->dev_audio = register_sound_dsp(&nsc_audio_fops, -1);
	
	nsc = &card->channels[0];

	OS_DbgMsg(KERN_INFO "geodeoss: Configuring %s found at IO 0x%04X IRQ %d\n", card_names[card_type],iobase,card->irq);
	pci_read_config_dword(pcidev, PCI_SUBSYSTEM_VENDOR_ID, &n);
	OS_DbgMsg(KERN_INFO "geodeoss:  subvendor id: 0x%08x\n",n); 

	//
	// turn off power management unless:
	//	- the user explicitly asks for it
	//
	if (use_pm != 1)
		use_pm = 0;

	if(!use_pm) 
	{
		OS_DbgMsg(KERN_INFO "geodeoss: not attempting power management.\n");
	}
	else 
	{
		if(!parse_power(card,pcidev)) 
		{
			printk(KERN_WARNING "%s%s%s",
				"geodeoss: ",
				"disabling PM: ",
				"no PCI power management interface found.\n");
			use_pm = 0; // This is how we track the PM setup
		}
	}

	if(use_pm) {
		//
		// DEVEL_TODO: The power managment register should probably
		// impact what happens next, instead of just being read into
		// a local variable and forgotten. Note that the callback is a
		// "veto machine" -- as no PM is implemented enabling PM prevent
		// "PM_SUSPEND" by a return a non-zero value which is read by
		//  the PM system as a "veto" of the state change.  This does
		//  not impact system shutdown.
		pci_read_config_dword(pcidev, card->power_regs, &n);
		OS_DbgMsg(KERN_INFO
			"geodeoss: PCI power management capability: 0x%x\n",n>>16);
		pmdev = pm_register(PM_PCI_DEV, PM_PCI_ID(pcidev), geode_pm_callback);
		if (pmdev)
			pmdev->data = card;
		else {
			printk(KERN_WARNING "%s%s%s",
				"geodeoss: ", "disabling PM: ",
				"PM registration failed\n");
			use_pm = 0; // PM setup failed
		}
	}

	card->mix.supported_mixers	= AC97_SUPPORTED_MASK;
	card->mix.stereo_mixers		= AC97_STEREO_MASK;
	card->mix.record_sources	= AC97_RECORD_MASK;
//	card->mix.read_mixer		= ac97_read_mixer; 
	card->mix.write_mixer		= ac97_write_mixer;
	card->mix.recmask_io		= ac97_recmask_io;
	card->mix.supported_mixers &= ~(SOUND_MASK_BASS|SOUND_MASK_TREBLE);

	if ((card->dev_mixer = register_sound_mixer(&nsc_mixer_fops, -1)) < 0) 
	{
		OS_DbgMsg("geodeoss: couldn't register mixer!\n");
	} 
	else 
	{
		memcpy(card->mix.mixer_state,mixer_defaults,sizeof(card->mix.mixer_state));
		mixer_push_state(card);
	}

	OS_DbgMsg(KERN_INFO "geodeoss: card->irq %d,\n", card->irq);
	if((ret=request_irq(card->irq, nsc_interrupt, SA_SHIRQ, card_names[card_type], card)))
	{
		OS_DbgMsg(KERN_ERR "geodeoss: unable to allocate irq %d,\n", card->irq);
		unregister_sound_mixer(card->dev_mixer);

		{
			struct nsc_state *s = &card->channels[0];

			if(s->dev_audio != -1)
				unregister_sound_dsp(s->dev_audio);
		}

		if (card_type==TYPE_GX2)
		{
			release_region(card->iobase, 128);		
		}

		unregister_reboot_notifier(&geode_nb);
		kfree(card);

		return ret;
	}

	pci_set_drvdata(pcidev,card);

	return 0;
}

static void geode_remove(struct pci_dev *pcidev) 
{
	struct nsc_card *card = pci_get_drvdata(pcidev);
	struct nsc_state *nsc = &card->channels[0];
	
	OS_DbgMsg("gxossaud: geode_remove\n");
	free_irq(card->irq, card);
	unregister_sound_mixer(card->dev_mixer);


	if(nsc->dev_audio != -1)
		unregister_sound_dsp(nsc->dev_audio);

	// geode_power(card,ACPI_D3);

	//
	//	Only Free the resources if they were allocated
	//
	if (gAllocatedIOResources)
	 	release_region(card->iobase, 128);

	kfree(card);
	pci_set_drvdata(pcidev,NULL);
}

static struct pci_device_id geode_pci_tbl[] __devinitdata = 
{
	{PCI_VENDOR_CYRIX,	PCI_DEVICE_ID_NSC_GX1,		PCI_ANY_ID, PCI_ANY_ID, 0, 0, TYPE_GX1},
	{PCI_VENDOR_NSC,	PCI_DEVICE_ID_NSC_SCXX00,	PCI_ANY_ID, PCI_ANY_ID, 0, 0, TYPE_SCXX00},
	{PCI_VENDOR_NSC,	PCI_DEVICE_ID_NSC_GX2,		PCI_ANY_ID, PCI_ANY_ID, 0, 0, TYPE_GX2},
	{0,}
};

MODULE_DEVICE_TABLE(pci, geode_pci_tbl);

static struct pci_driver geode_pci_driver = 
{
	name:"geodeoss",
	id_table:geode_pci_tbl,
	probe:geode_probe,
	remove:geode_remove,
};

int __init init_geode(void)
{
	int rc;

	OS_DbgMsg("geodeoss: init_geode\n");

	rc = pci_module_init(&geode_pci_driver);

	if (rc < 0)
		return rc;

	OS_DbgMsg("geodeoss: pci_module_init Succeded\n");

	if (register_reboot_notifier(&geode_nb))
		OS_DbgMsg(KERN_WARNING "geodeoss: reboot notifier registration failed; may not reboot properly.\n");

#ifdef MODULE
	printk(version);
	printk(KERN_INFO "geodeoss: use_pm=%d\n",use_pm);
#endif

	return 0;
}

static int geode_notifier(struct notifier_block *nb, unsigned long event, void *buf)
{
	// this notifier is called when the kernel is really shut down. 
	OS_DbgMsg("geodeoss: shutting down\n");
	// this will remove all card instances too 
	pci_unregister_driver(&geode_pci_driver);
	// XXX dunno about power management 
	return NOTIFY_OK;
}

void cleanup_geode(void) 
{
	printk("geodeoss: unloading\n");
	pci_unregister_driver(&geode_pci_driver);
	if(use_pm)
		pm_unregister_all(geode_pm_callback);
	unregister_reboot_notifier(&geode_nb);
}

int geode_pm_callback(struct pm_dev *dev, pm_request_t rqst, void *data) 
{
	struct nsc_card *card = (struct nsc_card*) dev->data;

	int retCode = 0;
	if ( ! card ) goto out;

	OS_DbgMsg("geodeoss: pm event 0x%x received for card %p\n", rqst, card);

	switch (rqst) 
	{
		case PM_SUSPEND: 
			printk(KERN_INFO " geodeoss: PM_SUSPEND\n");
			retCode = -EBUSY;
			// geode_suspend(card);
		break;
		case PM_RESUME: 
			printk(KERN_INFO " geodeoss: PM_RESUME\n");
			// geode_resume(card);
		break;
			printk(KERN_INFO " geodeoss: Unknown PM mode %d\n",rqst);
		//
		// We'd also like to find out about
		// power level changes because some biosen
		// do mean things to the geode when they
		// change their power state.
		//
        }
out:
	if (retCode)
		printk(KERN_WARNING " geodeoss: vetoing PM request\n");
	return retCode;
}

module_init(init_geode);
module_exit(cleanup_geode);
