#ifndef __NSCM_H 
#define __NSCM_H 
 
#include <linux/modversions.h>

#include "pcm.h" 
#include "mixer.h" 
#include "mpu401.h" 
#include "ac97_codec.h" 
#include "driver.h" 
#include "ainstr_simple.h" 
 
// 
// National's Device and Vendor IDs 
// 
 
#define PCI_VENDOR_ID_NSC		0x100B 
#define PCI_VENDOR_ID_CYRIX		0x1078 
	 
#define PCI_DEVICE_ID_UNICORN	0x0103 
#define PCI_DEVICE_ID_CENTAURUS	0x0103 
#define PCI_DEVICE_ID_SCORPIUS	0x0403 
#define PCI_DEVICE_ID_SC1200	0x0503 
#define PCI_DEVICE_ID_CS5535	0x002E 

#define DAC_RUNNING		1 
#define ADC_RUNNING		2 
 
#define NSC_MODE_NONE		0x00 
#define NSC_MODE_PLAY		0x01 
#define NSC_MODE_REC		0x02 
 
// Midi  
#define NSC_MPU401_PORT		0x98 
 
struct snd_stru_geode; 
struct snd_stru_channel; 
struct snd_stru_channelp; 
struct snd_stru_channelc; 
struct snd_stru_NSCdma; 
 
typedef struct snd_stru_geode geode_t; 
typedef struct snd_stru_channelp NSCchanp_t;	// Playback Channel 
typedef struct snd_stru_channelc NSCchanc_t;	// Record Channel 
typedef struct snd_stru_NSCdma NSCdma_t; 

// APU use in the driver 
enum snd_enum_apu_type { 
	pcm_play, 
	pcm_capture, 
	pcm_rateconv, 
	wavetable, 
	free 
}; 
 
// DMA Hack! 
struct snd_stru_NSCdma { 
	geode_t *card; 
	char *buf; 
	int size; 
 
	NSCdma_t *next; 
}; 
 
// Playback Channel 
struct snd_stru_channelp { 
	geode_t *card; 
	int num; 
 
	spinlock_t reg_lock; 
 
	unsigned int transs;	// Transfer Size 
	unsigned int transf;	// Transfer Fragment 
 
	// PCM stuff 
	volatile unsigned int mode; 
	unsigned char fmt; 
	volatile unsigned char enable; 
 
	unsigned long * PRDarray; 
	int		NumberOfPRDItems; 
	wait_queue_head_t interrupt_sleeper; 
}; 
 
// Record Channel 
struct snd_stru_channelc { 
	geode_t *card; 
	int num; 
 
	spinlock_t reg_lock; 
 
	unsigned int transs;	// Transfer Size 
	unsigned int transf;	// Transfer Fragment 
 
	// PCM stuff 
	unsigned int mode; 
	unsigned char fmt;
	volatile unsigned char enable; 
 
	unsigned long * PRDarray; 
	int		LastFragmentRecorded; 
	int		NumberOfPRDItems; 
	wait_queue_head_t interrupt_sleeper; 
}; 
 
struct snd_stru_geode { 
	// Module Config 
	int midi; 
	int gesbuf; 
	int pcmp, pcmc; 
 
	// Resources... 
	snd_dma_t *dma1ptr; 
	snd_irq_t *irqptr; 
 
	unsigned long io_port; 
 
	int type; 
	struct pci_dev *pci; 
	snd_card_t *card; 
	snd_pcm_t *pcm; 
 
	ac97_t *ac97; 
	snd_kmixer_t *mixer; 
	snd_rawmidi_t *rmidi; 
 
	spinlock_t reg_lock; 
	snd_info_entry_t *proc_entry; 
}; 
 
int snd_geode_create(snd_card_t * card, struct pci_dev *pci, 
					 snd_dma_t * dma1ptr, snd_irq_t * irqptr,  
					 int midi_enable, int gesbuf, int pcmp, int pcmc,  
                     geode_t ** resm); 
 
int snd_geode_free(geode_t * esm); 
void snd_geode_interrupt(geode_t * esm); 
 
int snd_geode_pcm(geode_t * esm, int device, snd_pcm_t ** rpcm); 
int snd_geode_mixer(geode_t * esm, int device, snd_pcm_t * pcm, snd_kmixer_t ** rmixer); 
int snd_geode_midi(geode_t * esm, int device, snd_rawmidi_t **); 
 
#endif				// __NSCM_H__ 
   