/* NSC_LIC_BSD
 *
 * Revision 2.1.6
 *
 * National Semiconductor Corporation Open Source License for 
 *
 *    Geode Linux ALSA audio driver
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c) August, 2002
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 *   * Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer. 
 *
 *   * Redistributions in binary form must reproduce the above 
 *     copyright notice, this list of conditions and the following 
 *     disclaimer in the documentation and/or other materials provided 
 *     with the distribution. 
 *
 *   * Neither the name of the National Semiconductor Corporation nor 
 *     the names of its contributors may be used to endorse or promote 
 *     products derived from this software without specific prior 
 *     written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER 
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN 
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF 
 * YOUR JURISDICTION. It is licensee's responsibility to comply with 
 * any export regulations applicable in licensee's jurisdiction. Under 
 * CURRENT (August 2002) U.S. export regulations this software 
 * is eligible for export from the U.S. and can be downloaded by or 
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed 
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran, 
 * Syria, Sudan, Afghanistan and any other country to which the U.S. 
 * has embargoed goods and services. 
 *
 * END_NSC_LIC_BSD */

#define __SND_OSS_COMPAT__ 
#define SND_MAIN_OBJECT_FILE 
#include "../../include/driver.h" 
#include "../../include/info.h" 
#include "../../include/geode.h" 

#include "os_inc.c"
#include "duraudio.c" 

PDURAUDIO	pGeode;

void		DumpLastBuffers (unsigned char *, int, int); 

//---------------------------------- 
//       Low Level Functions 
//---------------------------------- 
static void snd_geode_codec_init(void *private_data, ac97_t *ac97)
{
	//
	//   Set all the power state bits to 0 (Reg 26h)
	//
	DURAUDIO_CodecWrite ( pGeode, 0x26, 0x0000);
}

static void snd_geode_codec_write(void *private_data, unsigned short reg, unsigned short val) 
{ 
	if (reg) 
	{ 
		DURAUDIO_CodecWrite( pGeode, (unsigned char) reg, val); 
	} 

	udelay(100);
} 
 
static unsigned short snd_geode_codec_read(void *private_data, unsigned short reg) 
{ 
	return(DURAUDIO_CodecRead(pGeode, (unsigned char) reg)); 
}

//------------------
//  ALSA Interface  
//------------------
 
static int snd_geode_playback_ioctl(void *private_data, snd_pcm_subchn_t * subchn, unsigned int cmd, unsigned long *arg) 
{ 
/*
	int result;
	OS_DbgMsg ("*** snd_geode_playback_ioctl *** - CMD: %08X\n", cmd); 

	result = snd_pcm_lib_ioctl(private_data, subchn, cmd, arg);

	if (result < 0)
		return result;
*/
	return snd_pcm_lib_ioctl(private_data, subchn, cmd, arg); 
} 
 
static int snd_geode_capture_ioctl(void *private_data, snd_pcm_subchn_t * subchn, unsigned int cmd, unsigned long *arg) 
{ 
	OS_DbgMsg ("*** snd_geode_capture_ioctl ***\n"); 
 
	//
	// We calculate the Realrate later. So we don't need any IOCTL 
	//
	return snd_pcm_lib_ioctl(private_data, subchn, cmd, arg); 
} 
 
static int snd_geode_playback_prepare(void *private_data, snd_pcm_subchn_t * subchn) 
{ 
	snd_pcm_runtime_t *runtime = subchn->runtime; 
	NSCchanp_t *ns = snd_magic_cast(NSCchanp_t, runtime->private_data, -ENXIO); 
 
	OS_DbgMsg ("*** snd_geode_playback_prepare ***\n"); 

	ns->transs = snd_pcm_lib_transfer_size(subchn); 
	ns->transf = snd_pcm_lib_transfer_fragment(subchn); 

	init_waitqueue_head(&ns->interrupt_sleeper);

	OS_DbgMsg("rate: %08X\n", (unsigned short) runtime->format.rate);
	DURAUDIO_SetCodecRate(pGeode, CHANNEL0_PLAYBACK, (unsigned long) runtime->format.rate);

	OS_DbgMsg("Rate:   %d\n", runtime->format.rate);
	OS_DbgMsg("Format: %d\n", runtime->format.format);
	OS_DbgMsg("Voices: %d\n", runtime->format.voices);

	if (!DURAUDIO_AllocatePRD (pGeode, CHANNEL0_PLAYBACK, (unsigned long) ns->transs))
	{
		OS_DbgMsg("Failed allocating the PRD.\n");
		return -EINVAL;
	}

	// 
	//  Clean the buffer 
	// 
	memset (runtime->dma_area->buf, 0, ns->transs); 	
	
	DURAUDIO_InitPRD	 (pGeode, CHANNEL0_PLAYBACK, virt_to_bus(runtime->dma_area->buf));
	DURAUDIO_SetupPRD	 (pGeode, CHANNEL0_PLAYBACK, runtime->format.rate, (unsigned long) ns->transf, INTERVAL_BYTES);
	DURAUDIO_ClearStat	 (pGeode, CHANNEL0_PLAYBACK);						

	return 0; 
} 
 
static int snd_geode_capture_prepare(void *private_data, snd_pcm_subchn_t * subchn) 
{ 
 	snd_pcm_runtime_t *runtime = subchn->runtime; 
	NSCchanc_t *ns = snd_magic_cast(NSCchanc_t, runtime->private_data, -ENXIO); 

	OS_DbgMsg ("*** snd_geode_capture_prepare ***\n"); 

	ns->transs = snd_pcm_lib_transfer_size(subchn); 
	ns->transf = snd_pcm_lib_transfer_fragment(subchn); 

	init_waitqueue_head(&ns->interrupt_sleeper);

	OS_DbgMsg("rate: %d\n", (unsigned short) runtime->format.rate);
	DURAUDIO_SetCodecRate(pGeode, CHANNEL1_RECORD, (unsigned long) runtime->format.rate);

	if (!DURAUDIO_AllocatePRD (pGeode, CHANNEL1_RECORD, (unsigned long) ns->transs))
	{
		OS_DbgMsg("Failed allocating the PRD.\n");
		return 1;
	}

	DURAUDIO_InitPRD	 (pGeode, CHANNEL1_RECORD, virt_to_bus(runtime->dma_area->buf));
	DURAUDIO_SetupPRD	 (pGeode, CHANNEL1_RECORD, runtime->format.rate, (unsigned long) ns->transf, INTERVAL_BYTES);
	DURAUDIO_ClearStat	 (pGeode, CHANNEL1_RECORD);						

	return 0; 
} 

static int snd_geode_playback_trigger(void *private_data, snd_pcm_subchn_t * subchn, int up) 
{ 
	NSCchanp_t *ns = snd_magic_cast(NSCchanp_t, subchn->runtime->private_data, -ENXIO); 
	snd_pcm_runtime_t *runtime = subchn->runtime; 

	OS_DbgMsg ("*** snd_geode_playback_trigger ***\n"); 
 
	if (up) 
	{ 
		OS_DbgMsg("\n*\n**\n*** START (PLAY) ***\n"); 

		ns->enable |= DAC_RUNNING; 
		ns->mode |= NSC_MODE_PLAY; 

		//DURAUDIO_DisplayPRD (pGeode, CHANNEL0_PLAYBACK);

		DURAUDIO_StartDMA (pGeode, CHANNEL0_PLAYBACK); 
	} 
	else 
	{ 
		OS_DbgMsg ("*** STOP (PLAY) ***\n"); 
		DURAUDIO_StopDMA (pGeode, CHANNEL0_PLAYBACK);
		memset (runtime->dma_area->buf, 0x00, ns->transs); 
 	} 
 
	return 0; 
} 
 
static int snd_geode_capture_trigger(void *private_data, snd_pcm_subchn_t * subchn, int up) 
{ 
	NSCchanc_t *ns = snd_magic_cast(NSCchanc_t, subchn->runtime->private_data, -ENXIO); 
	snd_pcm_runtime_t *runtime = subchn->runtime; 

	OS_DbgMsg ("*** snd_geode_capture_trigger ***\n"); 
 
	if (up) 
	{ 
		OS_DbgMsg("*** START (REC) ***\n"); 

		ns->enable |= ADC_RUNNING; 
		ns->mode |= NSC_MODE_REC; 

		DURAUDIO_StartDMA(pGeode, CHANNEL1_RECORD); 
	} 
	else 
	{ 
		OS_DbgMsg ("*** STOP (PLAY) *** ns->mode:%08X\n", ns->mode); 
		DURAUDIO_StopDMA (pGeode, CHANNEL1_RECORD);
		memset (runtime->dma_area->buf, 0x00, ns->transs); 
 	} 
	return 0; 
} 
 
static unsigned int snd_geode_playback_pointer(void *private_data, snd_pcm_subchn_t * subchn) 
{ 
	OS_DbgMsg ("*** snd_geode_playback_pointer ***\n"); 

	return DURAUDIO_GetPosition (pGeode, CHANNEL0_PLAYBACK); 
} 
 
static unsigned int snd_geode_capture_pointer(void *private_data, snd_pcm_subchn_t * subchn) 
{ 
	OS_DbgMsg ("*** snd_geode_capture_pointer ***\n"); 

	return DURAUDIO_GetPosition (pGeode, CHANNEL1_RECORD); 
} 
 
static snd_pcm_hardware_t snd_geode_playback = { 
	SND_PCM_CHNINFO_MMAP |  
	SND_PCM_CHNINFO_BLOCK | SND_PCM_CHNINFO_INTERLEAVE | 
	SND_PCM_CHNINFO_BLOCK_TRANSFER | SND_PCM_CHNINFO_MMAP_VALID,	// flags  
	SND_PCM_FMT_S16_LE,												// hardware formats  
	SND_PCM_RATE_CONTINUOUS | SND_PCM_RATE_8000_48000,				// supported Rate  
	8000,	// Min. rate  
	48000,	// Max. rate 
	2,		// Min. voices  
	2,		// Max. voices  
	4096,	// Min. fragment size  
	32768,	// Max. fragment size  
	7,		// Fragment align  
	0,		// FIFO size (unknown)  
	8,		// Transfer block size  
	snd_geode_playback_ioctl, 
	snd_geode_playback_prepare, 
	snd_geode_playback_trigger, 
	snd_geode_playback_pointer 
}; 
 
static snd_pcm_hardware_t snd_geode_capture = { 
	// Matze: No mmap() for record at the moment  
	SND_PCM_CHNINFO_MMAP |  
	SND_PCM_CHNINFO_BLOCK |  SND_PCM_CHNINFO_INTERLEAVE | 
	SND_PCM_CHNINFO_BLOCK_TRANSFER| SND_PCM_CHNINFO_MMAP_VALID,
	SND_PCM_FMT_S16_LE,									// hardware formats 
	SND_PCM_RATE_CONTINUOUS | SND_PCM_RATE_8000_48000,	// supported Rate  
	8000,	// Min. rate  
	48000,	// Max. rate  
	2,		// Min. voices  
	2,		// Max. voices  
	4096,	// Min. fragment size  
	32768,	// Max. fragment size  
	7,		// Fragment align  
	0,		// FIFO size (unknown)  
	8,		// Transfer block size  
	snd_geode_capture_ioctl, 
	snd_geode_capture_prepare, 
	snd_geode_capture_trigger,  
	snd_geode_capture_pointer 
}; 
 
static void snd_geode_pcmp_free_subchn(void *private_data) 
{ 
	NSCchanp_t *ns = snd_magic_cast(NSCchanp_t, private_data, ); 

	OS_DbgMsg("snd_geode_pcmp_free_subchn\n");

	if (ns)  
	{ 
		DURAUDIO_FreePRD(pGeode, CHANNEL0_PLAYBACK);

		snd_magic_kfree(ns); 
	} 
} 
 
static void snd_geode_pcmc_free_subchn(void *private_data) 
{ 
	NSCchanc_t *ns = snd_magic_cast(NSCchanc_t, private_data, ); 
 
	OS_DbgMsg("snd_geode_pcmc_free_subchn\n");

	if (ns)  
	{ 
		DURAUDIO_FreePRD(pGeode, CHANNEL1_RECORD);

		snd_magic_kfree(ns); 
	} 
} 
  
static int snd_geode_playback_open(void *private_data, snd_pcm_subchn_t * subchn) 
{ 
	snd_pcm_runtime_t *runtime = subchn->runtime; 
	geode_t *nsc = snd_magic_cast(geode_t, private_data, -ENXIO); 
	NSCchanp_t *ns; 
	int err; 
 
	OS_DbgMsg("snd_geode_playback_open!\n"); 
 
	ns = snd_magic_kcalloc(NSCchanp_t, 0, GFP_KERNEL); 
 
	if ((err = snd_pcm_dma_alloc(subchn, nsc->dma1ptr, "geode - DAC")) < 0) 
		return err; 
 
	if (!ns)  
	{ 
		OS_DbgMsg("snd_geode_playback_open: Mem Error!\n"); 
		snd_pcm_dma_free(subchn); 
		return -ENOMEM; 
	} 
 
	ns->card = nsc; 
	ns->enable = 0; 
 
	runtime->private_data = ns; 
	runtime->private_free = snd_geode_pcmp_free_subchn; 
	runtime->hw = &snd_geode_playback; 
 
	snd_pcm_set_mixer(subchn, nsc->mixer->device, nsc->ac97->me_playback); 
 
	return 0; 
} 
 
static int snd_geode_capture_open(void *private_data, snd_pcm_subchn_t * subchn) 
{ 
	snd_pcm_runtime_t *runtime = subchn->runtime; 
	geode_t *nsc = snd_magic_cast(geode_t, private_data, -ENXIO); 
	NSCchanc_t *ns; 
	int err; 
 
	OS_DbgMsg("snd_geode_capture_open!\n"); 
 
	ns = snd_magic_kcalloc(NSCchanc_t, 0, GFP_KERNEL); 

	if ((err = snd_pcm_dma_alloc(subchn, nsc->dma1ptr, "geode - ADC")) < 0) 
		return err; 

	if (!ns)  
	{ 
		OS_DbgMsg("snd_capture_open: Mem Error!\n"); 
		snd_pcm_dma_free(subchn); 
		return -ENOMEM; 
	} 
 
	ns->card = nsc; 
	ns->enable = 0; 
 
	runtime->private_data = ns; 
	runtime->private_free = snd_geode_pcmc_free_subchn; 
	runtime->hw = &snd_geode_capture; 
 
	snd_pcm_set_mixer(subchn, nsc->mixer->device, nsc->ac97->me_capture); 
 
	return 0; 
} 
 
static int snd_geode_playback_close(void *private_data, snd_pcm_subchn_t * subchn) 
{ 
	OS_DbgMsg("snd_geode_playback_close!\n"); 
	snd_pcm_dma_free(subchn); 

	if (pGeode->AudioChannel[CHANNEL0_PLAYBACK].PRDTable)
	{
		DURAUDIO_FreePRD(pGeode, CHANNEL0_PLAYBACK);
	}
	return 0; 
} 
 
static int snd_geode_capture_close(void *private_data, snd_pcm_subchn_t * subchn) 
{ 
	OS_DbgMsg("snd_geode_capture_close!\n"); 
	snd_pcm_dma_free(subchn); 

	if (pGeode->AudioChannel[CHANNEL1_RECORD].PRDTable)
	{
		DURAUDIO_FreePRD(pGeode, CHANNEL1_RECORD);
	}
	return 0; 
} 
 
static void snd_geode_pcm_free(void *private_data) 
{ 
	geode_t *nsc = snd_magic_cast(geode_t, private_data, ); 

	nsc->pcm = NULL; 
} 
 
int snd_geode_pcm(geode_t * nsc, int device, snd_pcm_t ** rpcm) 
{ 
	snd_pcm_t *pcm; 
	int err; 
 
	*rpcm = NULL; 
 
	if ((err = snd_pcm_new(nsc->card, "NSC geode", device, nsc->pcmp, nsc->pcmc, &pcm)) < 0) 
		return err; 
 
	pcm->private_data = nsc; 
	pcm->private_free = snd_geode_pcm_free; 
 
	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].private_data = nsc; 
	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].open = snd_geode_playback_open; 
	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].close = snd_geode_playback_close; 
 
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].private_data = nsc; 
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].open = snd_geode_capture_open; 
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].close = snd_geode_capture_close; 
 
	if (nsc->pcmc > 0) 
	{ 
		pcm->info_flags = SND_PCM_INFO_PLAYBACK |  
						  SND_PCM_INFO_CAPTURE  | 
						  SND_PCM_INFO_DUPLEX; 
	} 
	else 
		pcm->info_flags = SND_PCM_INFO_PLAYBACK; 
 
	strcpy(pcm->name, "NSC geode"); 
 
	*rpcm = nsc->pcm = pcm; 
 
	return 0; 
} 
 
//--------------------- 
//    Mixer Functions    
//--------------------- 
 
static void snd_geode_mixer_free_ac97(void *private_data) 
{ 
	geode_t * nsc = snd_magic_cast(geode_t, private_data, ); 
	nsc->mixer = NULL; 
} 
 
int snd_geode_mixer(geode_t * nsc, int device, snd_pcm_t * pcm, snd_kmixer_t ** rmixer) 
{ 
	ac97_t ac97; 
	snd_kmixer_t *mixer; 
	int err; 
 
	*rmixer = NULL; 
 
	memset(&ac97, 0, sizeof(ac97)); 
	ac97.write = snd_geode_codec_write; 
	ac97.read = snd_geode_codec_read; 
	ac97.private_data = nsc; 
	ac97.private_free = snd_geode_mixer_free_ac97; 
	ac97.init = snd_geode_codec_init; 
 
	if ((err = snd_ac97_mixer(nsc->card, device, &ac97, 1, &pcm->device, &mixer)) < 0) 
	{ 
		OS_DbgMsg("****  FAILED ON snd_ac97_mixer ****\n"); 
		return err; 
	} 
 
	nsc->ac97 = snd_magic_cast(ac97_t, (mixer->private_data), -ENXIO); 
	*rmixer = nsc->mixer = mixer; 
	return 0; 
} 

//------------------------------- 
//   Config & Create Soundcard  
//------------------------------- 
 
int snd_geode_free(geode_t * nsc) 
{ 
	snd_magic_kfree(nsc); 
	return 0; 
} 
 
int snd_geode_create(snd_card_t * card,  
					  struct pci_dev *pci,  
					  snd_dma_t * dma1ptr,  
					  snd_irq_t * irqptr,  
					  int midi_enable,  
					  int gesbuf,  
					  int pcmp,  
					  int pcmc,  
					  geode_t ** rnsc) 
{ 
	geode_t *nsc; 
	int err; 
 
 
	static snd_device_ops_t ops = { 
		(snd_dev_free_t *)snd_geode_free, 
		NULL, 
		NULL 
	}; 
 
	*rnsc = NULL; 
 
	nsc = (geode_t *) snd_magic_kcalloc(geode_t, 0, GFP_KERNEL); 
	if (!nsc) 
		return -ENOMEM; 
 
	nsc->type = card->type; 
	spin_lock_init(&nsc->reg_lock); 
	nsc->card = card; 
	nsc->pci = pci; 
	nsc->irqptr = irqptr; 
	nsc->dma1ptr = dma1ptr; 
	nsc->midi = midi_enable; 
	nsc->gesbuf = gesbuf; 
	nsc->pcmp = pcmp; 
	nsc->pcmc = pcmc; 
 
	nsc->io_port = pci_resource_start(pci, 0); 
 
	OS_DbgMsg("IO:  0x%lx\n", nsc->io_port); 
	OS_DbgMsg("IRQ: 0x%lx\n", nsc->irqptr->irq); 
 
	pci_set_master(pci); 
 
	if (nsc->irqptr->irq)
	{
		pGeode=DURAUDIO_Initialize((unsigned int) nsc->irqptr->irq);	
	}
	else
	{
		pGeode=DURAUDIO_Initialize(5);	//  If the routing table is bad...
	}

	if (!pGeode)   
	{ 
		OS_DbgMsg("GEODE: couldn't Initialize the Audio device!\n"); 
		return -1; 
	} 

	if ((err = snd_device_new(card, SND_DEV_LOWLEVEL, nsc, 0, &ops, NULL)) < 0)  
	{ 
		snd_geode_free(nsc); 
		return err; 
	} 
 
	*rnsc = nsc; 
 
	return 0; 
} 
 
//--------------- 
//    Midi Part  
//--------------- 
int snd_geode_midi(geode_t * nsc, int device, snd_rawmidi_t ** rawmidi) 
{ 
	int mpu_port; 
	int err; 
	snd_rawmidi_t *rm; 
 
	rm = NULL; 
 
	mpu_port = nsc->io_port + NSC_MPU401_PORT; 
 
	err = snd_mpu401_uart_new(nsc->card, device, 
				  MPU401_HW_MPU401, 
				  mpu_port, nsc->irqptr->irq, &rm); 
	if (err < 0) 
		return err; 
 
	*rawmidi = nsc->rmidi = rm; 
	return 0; 
} 
 
//----------------- 
//  IRQ Handling  
//----------------- 
static void nsc_update_ptrp(snd_pcm_subchn_t * subchn) 
{ 
	NSCchanp_t *ns; 

	//OS_DbgMsg("*** nsc_update_ptrp ***\n"); 

	if (subchn->runtime && subchn->runtime->private_data)  
	{ 		
		//OS_DbgMsg ("*** INT - PLAY ***\n"); 

 		ns = snd_magic_cast(NSCchanp_t, subchn->runtime->private_data, ); 

		if (ns->mode & NSC_MODE_PLAY)
		{ 
		 	DURAUDIO_ClearStat(pGeode, CHANNEL0_PLAYBACK);		// Clear the Playback Status bit 

			snd_pcm_transfer_done(subchn);	 
		} 
		else
		{
			ns->enable &= ~DAC_RUNNING; 
			wake_up_interruptible (&ns->interrupt_sleeper);
			//OS_DbgMsg("*** EOT ***\n"); 
		}
	} 
} 

static void nsc_update_ptrc(snd_pcm_subchn_t * subchn) 
{ 
	NSCchanc_t *ns; 

	OS_DbgMsg("*** nsc_update_ptrc ***\n"); 

	if (subchn->runtime && subchn->runtime->private_data)  
	{ 		
		OS_DbgMsg ("*** INT - REC ***\n"); 

 		ns = snd_magic_cast(NSCchanc_t, subchn->runtime->private_data, ); 

		if (ns->mode & NSC_MODE_REC)
		{ 
		 	DURAUDIO_ClearStat(pGeode, CHANNEL1_RECORD);		// Clear the Recording Status bit 

			snd_pcm_transfer_done(subchn);	 
		} 
		else
		{
			ns->enable &= ~ADC_RUNNING; 
			wake_up_interruptible (&ns->interrupt_sleeper);
			OS_DbgMsg("*** EOT ***\n"); 
		}
	} 
} 

void snd_geode_interrupt(geode_t * nsc) 
{ 
	snd_pcm_subchn_t *subchn; 
	unsigned char    ID;

	ID=DURAUDIO_InterruptID(pGeode);

	if (ID & DMA0_INTERRUPT)
	{
		DURAUDIO_ClearIRQ (pGeode);
		subchn = nsc->pcm->chn[SND_PCM_CHANNEL_PLAYBACK].subchn; 
		nsc_update_ptrp(subchn); 
	}

	if (ID & DMA1_INTERRUPT)
	{
		DURAUDIO_ClearIRQ (pGeode);
		subchn = nsc->pcm->chn[SND_PCM_CHANNEL_CAPTURE].subchn; 
		nsc_update_ptrc(subchn);  
	}

} 

void DumpLastBuffers (unsigned char *buffer, int BlockSize, int FragSize) 
{ 
	int i, j, NumberOfFrags, counter; 
 
	counter=0; 
 
	OS_DbgMsg("*****   Buffer: %08X   BlockSize: %d   FragSize:%d    *****\n", (unsigned int) buffer,  BlockSize, FragSize); 
	 
	NumberOfFrags=BlockSize/FragSize; 
	for (i=0; i<2; ++i) 
	{ 
		OS_DbgMsg("*********  Fragment %d  *********\n", i); 
		OS_DbgMsg("00000 "); 

		for (j=0;j<FragSize;j++) 
		{ 
			OS_DbgMsg("%02X ", *(buffer+(i*FragSize)+j)); 
			if((counter++)>16) 
			{ 
				OS_DbgMsg("\n %05d ", (unsigned int) j); 
				counter=0; 
			} 
		} 
		counter=0; 
		OS_DbgMsg("\n"); 
	} 
} 
 
//----------------- 
//   Modul Stuff  
//----------------- 
 
EXPORT_SYMBOL(snd_geode_create); 
EXPORT_SYMBOL(snd_geode_free); 
EXPORT_SYMBOL(snd_geode_pcm); 
EXPORT_SYMBOL(snd_geode_mixer); 
EXPORT_SYMBOL(snd_geode_midi); 
EXPORT_SYMBOL(snd_geode_interrupt); 
 
static int __init alsa_geode_init(void) 
{ 
	return 0; 
} 
 
static void __exit alsa_geode_exit(void) 
{ 
} 
 
module_init(alsa_geode_init) 
module_exit(alsa_geode_exit) 
